from v0.relationalai.semantics.metamodel import ir, builtins, helpers, types
from v0.relationalai.semantics.metamodel.visitor import collect_by_type
from v0.relationalai.semantics.metamodel.util import FrozenOrderedSet
from v0.relationalai.semantics.lqp import ir as lqp, utils, types as lqp_types, builtins as lqp_builtins
from v0.relationalai.semantics.lqp.primitives import lqp_avg_op, lqp_operator, build_primitive
from v0.relationalai.semantics.lqp.pragmas import pragma_to_lqp_name
from v0.relationalai.semantics.lqp.types import meta_type_to_lqp
from v0.relationalai.semantics.lqp.constructors import (
    mk_abstraction, mk_and, mk_exists, mk_or, mk_pragma, mk_primitive,
    mk_specialized_value, mk_type, mk_value, mk_attribute
)
from v0.relationalai.semantics.lqp.utils import TranslationCtx, gen_unique_var
from v0.relationalai.semantics.lqp.validators import assert_valid_input
from v0.relationalai.semantics.lqp.rewrite.functional_dependencies import (
    normalized_fd, contains_only_declarable_constraints
)
from decimal import Decimal as PyDecimal
from datetime import datetime, date, timezone
from typing import Tuple, cast, Union, Optional
from warnings import warn
import re
import uuid

# Main access point for translating metamodel to lqp. Converts the model IR to an LQP epoch.
def to_lqp(model: ir.Model, fragment_name: bytes, ctx: TranslationCtx) -> tuple[Optional[tuple], lqp.Epoch]:
    assert_valid_input(model)
    decls: list[lqp.Declaration] = []
    reads: list[lqp.Read] = []

    # LQP only accepts logical tasks
    # These are asserted at init time
    root = cast(ir.Logical, model.root)
    for subtask in root.body:
        assert isinstance(subtask, ir.Logical)
        new_decls = _translate_to_decls(ctx, subtask)
        decls.extend(new_decls)

    # Add pyrel error attributes to reads.
    for err_id in _extract_pyrel_error_ids(ctx, model):
        ctx.output_ids.append(err_id)

    reads.extend(_get_output_reads(ctx.output_ids))

    export_info = None
    if len(ctx.export_ids) > 0:
        export_filename, col_types, export_reads = _get_export_reads(ctx.export_ids)
        reads.extend(export_reads)
        export_info = (export_filename, col_types)

    debug_info = lqp.DebugInfo(id_to_orig_name=ctx.rel_id_to_orig_name, meta=None)
    fragment_id = lqp.FragmentId(id=fragment_name, meta=None)
    fragment = lqp.Fragment(id=fragment_id, declarations=decls, meta=None, debug_info=debug_info)
    define_op = lqp.Define(fragment=fragment, meta=None)

    txn = lqp.Epoch(
        reads=reads,
        writes=[lqp.Write(write_type=define_op, meta=None)],
        meta=None
    )

    return (export_info, txn)

def _effect_bindings(effect: Union[ir.Output, ir.Update]) -> list[ir.Value]:
    if isinstance(effect, ir.Output):
        # Unions may not return anything. The generated IR contains a None value when this
        # happens. We ignore it here.
        return [v for v in helpers.output_values(effect.aliases) if v]
    else:
        return list(effect.args)

def _get_output_reads(output_ids: list[tuple[lqp.RelationId, str]]) -> list[lqp.Read]:
    reads = []
    for (rel_id, name) in output_ids:
        assert isinstance(rel_id, lqp.RelationId)
        output = lqp.Output(name=name, relation_id=rel_id, meta=None)
        reads.append(lqp.Read(read_type=output, meta=None))
    return reads

def _get_export_reads(export_ids: list[tuple[lqp.RelationId, int, lqp.Type]]) -> tuple[str, list, list[lqp.Read]]:
    reads = []
    csv_columns = []
    col_info = []
    for (col_id, col_num, col_type) in sorted(export_ids, key=lambda x: x[1]):
        col_name = f"col{col_num:03}"
        csv_columns.append(lqp.ExportCSVColumn(column_name=col_name, column_data=col_id, meta=None))
        col_info.append((col_name, col_type))

    # Generate a random name for the internal export path
    export_filename = "export_" + str(uuid.uuid4()).replace("-", "_")

    # Note that the engine will append the transaction id to the export path for access
    # control reasons. So the actual final path will be
    # `.../export/{export_filename}/data_{txn_id}`
    export_path = f"snowflake://APP_STATE.RAI_INTERNAL_STAGE/export/{export_filename}/data"
    export_csv_config = lqp.ExportCSVConfig(
        path=export_path,
        data_columns=csv_columns,
        compression="gzip",
        partition_size=200,
        syntax_escapechar='"',  # To follow Snowflake's expected format
        meta=None,
    )
    reads.append(lqp.Read(read_type=lqp.Export(config=export_csv_config, meta=None), meta=None))
    return (export_filename, col_info, reads)

def _translate_to_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp.Declaration]:
    if contains_only_declarable_constraints(rule):
        return _translate_to_constraint_decls(ctx, rule)
    else:
        return _translate_to_standard_decl(ctx, rule)

def _translate_to_constraint_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp.Declaration]:
    constraint_decls: list[lqp.Declaration] = []
    for task in rule.body:
        assert isinstance(task, ir.Require)
        fd = normalized_fd(task)
        assert fd is not None

        # check for unresolved types
        if any(types.is_any(var.type) for var in fd.keys + fd.values):
            warn(f"Ignoring FD with unresolved type: {fd}")
            continue

        lqp_typed_keys = [_translate_term(ctx, key) for key in fd.keys]
        lqp_typed_values = [_translate_term(ctx, value) for value in fd.values]
        lqp_typed_vars:list[Tuple[lqp.Var, lqp.Type]] = lqp_typed_keys + lqp_typed_values # type: ignore
        lqp_guard_atoms = [_translate_to_atom(ctx, atom) for atom in fd.guard]
        lqp_guard = mk_abstraction(lqp_typed_vars, mk_and(lqp_guard_atoms))
        lqp_keys:list[lqp.Var] = [var for (var, _) in lqp_typed_keys] # type: ignore
        lqp_values:list[lqp.Var] = [var for (var, _) in lqp_typed_values] # type: ignore

        fd_decl = lqp.FunctionalDependency(
            guard=lqp_guard,
            keys=lqp_keys,
            values=lqp_values,
            meta=None
        )

        constraint_decls.append(fd_decl)

    return constraint_decls

def _translate_to_standard_decl(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp.Declaration]:
    effects = collect_by_type((ir.Output, ir.Update), rule)
    aggregates = collect_by_type(ir.Aggregate, rule)
    ranks = collect_by_type(ir.Rank, rule)

    # TODO: should this ever actually come in as input?
    if len(effects) == 0:
        return []

    assert len(ranks) == 0 or len(aggregates) == 0, "rules cannot have both aggregates and ranks"

    conjuncts = []
    for task in rule.body:
        if isinstance(task, (ir.Output, ir.Update)):
            continue
        conjuncts.append(_translate_to_formula(ctx, task))

    # Aggregates reduce over the body
    if aggregates or ranks:
        aggr_body = mk_and(conjuncts)
        conjuncts = []
        for aggr in aggregates:
            conjuncts.append(_translate_aggregate(ctx, aggr, aggr_body))
        for rank in ranks:
            conjuncts.append(_translate_rank(ctx, rank, aggr_body))

    return [_translate_effect(ctx, effect, mk_and(conjuncts)) for effect in effects]

def _translate_annotations(annotations: FrozenOrderedSet[ir.Annotation]) -> list[lqp.Attribute]:
    attributes = []
    for annotation in annotations:

        if annotation.relation.name in lqp_builtins.annotations_to_emit:
            if any(not isinstance(a, ir.Literal) for a in annotation.args):
                warn("LQP currently ignores annotation parameters with non-literal values")
                continue

            # Convert literal arguments to LQP values
            args = []
            for a in annotation.args:
                assert isinstance(a, ir.Literal)
                args.append(mk_value(a.value))
            attributes.append(mk_attribute(annotation.relation.name, args))
    return attributes

# Translates an effect (export, output, or update) into the corresponding def. Note that
# this method only generates the def and not the LQP read operation, which is added later by
# `_get_output_reads` and `_get_export_reads`.
def _translate_effect(ctx: TranslationCtx, effect: Union[ir.Output, ir.Update], body: lqp.Formula) -> lqp.Declaration:
    bindings = _effect_bindings(effect)

    def _is_export(e):
        return isinstance(e, ir.Output) and builtins.export_annotation in e.annotations

    if isinstance(effect, ir.Output):
        projection, eqs, suffix = _translate_output_bindings(ctx, bindings)
        meta_id = effect.id

        if _is_export(effect):
            def_name = "export_relation" + suffix
        else:
            def_name = "output" + suffix

        def_name = ctx.output_names.get_name_by_id(meta_id, def_name)
        rel_id = get_output_id(ctx, def_name, meta_id)
    else:
        projection, eqs = _translate_bindings(ctx, bindings)
        def_name = effect.relation.name
        rel_id = get_relation_id(ctx, effect.relation, projection)

    eqs.append(body)
    new_body = mk_and(eqs)

    # Context bookkeeping for exports and outputs
    if _is_export(effect):
        # The row id is the first n-1 elements, and the actual data is the last element. Its
        # type is stored in the first element of the tuple.
        col_type = projection[-1][1]
        _col_num_match = re.search(r"export_relation_col([0-9]+)", def_name)
        assert _col_num_match, f"Could not find column number in suffix: {def_name}"
        col_num = int(_col_num_match.group(1))
        ctx.export_ids.append((rel_id, col_num, col_type))
    elif isinstance(effect, ir.Output):
        ctx.output_ids.append((rel_id, def_name))

    # First we collect annotations on the effect itself, e.g. from something like
    # `select(...).annotate(...)`.
    annotations = effect.annotations
    if isinstance(effect, ir.Update):
        # Then we translate annotations on the relation itself, e.g.
        # ```
        # Bar.foo = model.Relationship(...)
        # Bar.foo.annotate(...)
        # ```
        annotations = annotations | effect.relation.annotations

    return lqp.Def(
        name = rel_id,
        body = mk_abstraction(projection, new_body),
        attrs = _translate_annotations(annotations),
        meta = None,
    )

def _translate_output_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.Type]], list[lqp.Formula], str]:
    symbol_literals = []
    non_symbols = []
    for binding in bindings:
        if isinstance(binding, ir.Literal) and binding.type == types.Symbol:
            symbol_literals.append(binding.value)
        else:
            non_symbols.append(binding)
    projection, eqs = _translate_bindings(ctx, non_symbols)
    if len(symbol_literals) > 0:
        name_suffix = "_"
        name_suffix += "_".join(symbol_literals)
    else:
        name_suffix = ""

    return projection, eqs, name_suffix

def _translate_rank(ctx: TranslationCtx, rank: ir.Rank, body: lqp.Formula) -> lqp.Formula:
    # Ascending rank is constructed using rel_primitive_sort. If a limit is added to an
    # ascending rank we can use rel_primitive_top for an efficient evaluation.
    #
    # Descending rank is constructed as an ascending sort, then the ascending rank is
    # subtracted from a count of the elements (plus 1 so that we still start at 1).
    # Adding a limit to a descending rank is done by adding a filter of rank <= limit.

    # Limits are the sort plus a filter on rank <= limit.
    if all(o for o in rank.arg_is_ascending):
        ascending = True
    elif all(not o for o in rank.arg_is_ascending):
        ascending = False
    else:
        raise Exception("Mixed orderings in rank are not supported yet.")

    # Filter out the group-by variables, since they are introduced outside the rank.
    input_args, input_eqs = _translate_bindings(ctx, list(rank.args))
    introduced_meta_projs = [arg for arg in rank.projection if arg not in rank.group and arg not in rank.args]
    projected_args, projected_eqs = _translate_bindings(ctx, list(introduced_meta_projs))

    # rank expects an Int128 result, but the primitive will return an Int64 result.
    # we need to set up an intermediary variable to hold the Int64 result, and a cast
    # to convert it to Int128.
    result_var, _ = _translate_term(ctx, rank.result)
    # The primitive will return an Int64 result, so we need a var to hold the intermediary.
    result_64_var = gen_unique_var(ctx, "v_rank")
    result_64_type = mk_type(lqp.TypeName.INT)

    cast = lqp.Cast(input=result_64_var, result=result_var, meta=None)

    body = mk_and([body] + input_eqs + projected_eqs)
    abstr_args = input_args + projected_args

    if ascending:
        ranker = _translate_ascending_rank(ctx, rank.limit, result_64_var, body, abstr_args)
    else:
        ranker = _translate_descending_rank(ctx, rank.limit, result_64_var, body, abstr_args)

    return mk_exists([(result_64_var, result_64_type)], mk_and([ranker, cast]))

def _translate_descending_rank(ctx: TranslationCtx, limit: int, result: lqp.Var, body: lqp.Formula, abstr_args) -> lqp.Formula:
    result_var = result
    result_type = mk_type(lqp.TypeName.INT)

    # Rename abstracted args in the body to new variable names
    var_map = {var.name: gen_unique_var(ctx, 't_' + var.name) for (var, _) in abstr_args}
    body = utils.rename_vars_formula(body, var_map)
    new_abstr_args = [(var_map[var.name], typ) for (var, typ) in abstr_args]

    # Construct a conjunction of the ranking, a counter for the body, a subtraction
    # of the rank from the count and an addition of 1. Wrap this in an abstraction.
    count_res = gen_unique_var(ctx, "count_res")

    # Add one to the count to account for the rank starting at 1.
    one, one_eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "one")
    one_bigger = gen_unique_var(ctx, "one_bigger")
    addition = mk_primitive("rel_primitive_add_monotype", [count_res, one, one_bigger])

    # Subtract the rank from the count + 1
    asc_rank = gen_unique_var(ctx, "asc_rank")
    subtraction = mk_primitive("rel_primitive_subtract_monotype", [one_bigger, result_var, asc_rank])

    # Construct the ranking
    desc_ranking_terms = [asc_rank] + [v[0] for v in abstr_args]
    ranking = lqp.FFI(
        meta=None,
        name="rel_primitive_sort",
        args=[mk_abstraction(new_abstr_args, body)],
        terms=desc_ranking_terms,
    )

    # Count the number of rows in the body
    count_type = meta_type_to_lqp(types.Int64)
    count_var, count_eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "counter")
    desc_body = mk_and([body, count_eq])
    aggr_abstr_args = new_abstr_args + [(count_var, count_type)]
    count_aggr = lqp.Reduce(
        op=lqp_operator(
            ctx,
            "count",
            "count",
            mk_type(lqp.TypeName.INT)
        ),
        body=mk_abstraction(aggr_abstr_args, desc_body),
        terms=[count_res],
        meta=None
    )

    # Bring it all together and do the maths.
    ranking = mk_exists(
        vars=[
            (asc_rank, result_type),
            (count_res, result_type),
            (one, result_type),
            (one_bigger, result_type)
        ],
        value=mk_and([ranking, count_aggr, one_eq, addition, subtraction])
    )

    # If there is a limit, we need to add a filter to the ranking.
    # Wrap with a rank <= limit
    if limit != 0:
        limit_term, _ = _translate_term(ctx, ir.Literal(types.Int64, limit))
        limiter = mk_primitive("rel_primitive_lt_eq_monotype", [result_var, limit_term])
        ranking = mk_and([ranking, limiter])

    return ranking

def _translate_ascending_rank(ctx: TranslationCtx, limit: int, result_var: lqp.Var, body: lqp.Formula, abstr_args) -> lqp.Formula:
    terms = [result_var] + [v[0] for v in abstr_args]

    # Rename abstracted args in the body to new variable names
    var_map = {var.name: gen_unique_var(ctx, 't_' + var.name) for (var, _) in abstr_args}
    body = utils.rename_vars_formula(body, var_map)
    new_abstr_args = [(var_map[var.name], typ) for (var, typ) in abstr_args]
    sort_abstr = mk_abstraction(new_abstr_args, body)

    if limit == 0:
        return lqp.FFI(
            meta=None,
            name="rel_primitive_sort",
            args=[sort_abstr],
            terms=terms,
        )
    else:
        limit_type = meta_type_to_lqp(types.Int64)
        limit_var, limit_eq = constant_to_var(ctx, to_lqp_value(limit, types.Int64), "limit")
        limit_abstr = mk_abstraction([(limit_var, limit_type)], limit_eq)
        return lqp.FFI(
            meta=None,
            name="rel_primitive_top",
            args=[sort_abstr, limit_abstr],
            terms=terms,
        )

def _rename_shadowed_abstraction_vars(
    ctx: TranslationCtx,
    aggr: ir.Aggregate,
    abstr_args: list[Tuple[lqp.Var, lqp.Type]],
    body_conjs: list[lqp.Formula]
) -> list[Tuple[lqp.Var, lqp.Type]]:
    """
    Rename abstraction variables that shadow group-by variables.

    This can happen when the same variable appears in both aggr.group and as an input
    to the aggregation, e.g., min(Person.age).per(Person.age). The group-by variables
    are in the outer scope, while the abstraction parameters are in the inner scope,
    so we need different names to avoid shadowing.
    """
    # Get the LQP names of group-by variables
    group_var_names = set()
    for group_var in aggr.group:
        lqp_var = _translate_var(ctx, group_var)
        group_var_names.add(lqp_var.name)

    # Rename any abstraction parameters that conflict with group-by variables
    renamed_abstr_args = []
    for var, typ in abstr_args:
        if var.name in group_var_names:
            # This variable shadows a group-by variable, so rename it
            fresh_var = gen_unique_var(ctx, var.name)
            # Add an equality constraint: fresh_var == var
            # var is a free variable referring to the outer scope group-by variable
            body_conjs.append(mk_primitive("rel_primitive_eq", [fresh_var, var]))
            renamed_abstr_args.append((fresh_var, typ))
        else:
            renamed_abstr_args.append((var, typ))

    return renamed_abstr_args

def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp.Formula) -> Union[lqp.Reduce, lqp.Formula]:
    # TODO: handle this properly
    aggr_name = aggr.aggregation.name
    supported_aggrs = ("sum", "count", "avg", "min", "max", "rel_primitive_solverlib_ho_appl")
    assert aggr_name in supported_aggrs, f"only support {supported_aggrs} for now, not {aggr.aggregation.name}"

    meta_output_terms = []
    meta_input_terms = []

    for (field, arg) in zip(aggr.aggregation.fields, aggr.args):
        if field.input:
            meta_input_terms.append(arg)
        else:
            meta_output_terms.append(arg)

    output_terms = [_translate_term(ctx, term) for term in meta_output_terms]
    output_vars = [term[0] for term in output_terms]

    body_conjs = [body]
    input_args, input_eqs = _translate_bindings(ctx, meta_input_terms)

    # TODO: Can this safely be applied to all aggregates?
    if aggr_name in ("sum", "min", "max"):
        assert len(output_terms) == 1, f"{aggr_name} expects a single output variable"
        assert len(meta_input_terms) == 1, f"{aggr_name} expects a single input variable"
        assert isinstance(meta_output_terms[0], ir.Var)
        assert input_args[0][1] == output_terms[0][1], f"{aggr_name}({input_args[0][1].type_name}) had output type of {output_terms[0][1].type_name}"

    # Filter out the group-by variables, since they are introduced outside the aggregation.
    # Input terms are added later below.
    introduced_meta_projs = [arg for arg in aggr.projection if arg not in aggr.group and arg not in meta_input_terms]
    projected_args, projected_eqs = _translate_bindings(ctx, list(introduced_meta_projs))
    body_conjs.extend(input_eqs)
    body_conjs.extend(projected_eqs)
    abstr_args: list[Tuple[lqp.Var, lqp.Type]] = projected_args + input_args

    # Rename abstraction variables that shadow group-by variables
    abstr_args = _rename_shadowed_abstraction_vars(ctx, aggr, abstr_args, body_conjs)

    if aggr_name == "count":
        assert len(output_terms) == 1, "Count and avg expect a single output variable"
        assert isinstance(meta_output_terms[0], ir.Var)
        # Count sums up "1" for each row. We use the expected output type for the type
        # of the count variable.
        typ = meta_type_to_lqp(meta_output_terms[0].type)
        one_var, eq = constant_to_var(ctx, to_lqp_value(1, meta_output_terms[0].type), "one")
        body_conjs.append(eq)
        abstr_args.append((one_var, typ))

    # Average needs to wrap the reduce in Exists(Conjunction(Reduce, div))
    if aggr_name == "avg":
        assert len(output_vars) == 1, "avg should only have one output variable"
        output_var = output_vars[0]

        # Count sums up "1" for each row. We make the reasonably safe assumption that there
        # are less than 2^31 rows in the body.
        count_type = meta_type_to_lqp(types.Int64)
        one_var, eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "one")
        body_conjs.append(eq)
        abstr_args.append((one_var, count_type))

        # The average will produce two output variables: sum and count.
        sum_result = gen_unique_var(ctx, "sum")
        count_result = gen_unique_var(ctx, "count")

        # Second to last is the variable we're summing over.
        (sum_var, sum_type) = abstr_args[-2]
        body = mk_and(body_conjs)

        result = lqp.Reduce(
            op=lqp_avg_op(ctx, aggr.aggregation.name, sum_var.name, sum_type),
            body=mk_abstraction(abstr_args, body),
            terms=[sum_result, count_result],
            meta=None,
        )

        if sum_type == count_type:
            div = mk_primitive("rel_primitive_divide_monotype", [sum_result, count_result, output_var])
            conjunction = mk_and([result, div])

            # Finally, we need to wrap everything in an `exists` to project away the sum and
            # count variables and only keep the result of the division.
            result_terms = [(sum_result, sum_type), (count_result, count_type)]
        else:
            # If the sum type and count type don't match, we need to cast the count
            count_casted = gen_unique_var(ctx, "count_casted")
            count_cast = lqp.Cast(input=count_result, result=count_casted, meta=None)

            div = mk_primitive("rel_primitive_divide_monotype", [sum_result, count_casted, output_var])
            conjunction = mk_and([result, count_cast, div])

            # Finally, we need to wrap everything in an `exists` to project away the sum and
            # count variables and only keep the result of the division.
            result_terms = [(sum_result, sum_type), (count_result, count_type), (count_casted, sum_type)]

        return mk_exists(result_terms, conjunction)

    # `input_args` hold the types of the input arguments, but they may have been modified
    # if we're dealing with a count, so we use `abstr_args` to find the type.
    (aggr_arg, aggr_arg_type) = abstr_args[-1]
    body = mk_and(body_conjs)

    # Group-bys do not need to be handled at all, since they are introduced outside already
    reduce = lqp.Reduce(
        op=lqp_operator(ctx, aggr.aggregation.name, aggr_arg.name, aggr_arg_type),
        body=mk_abstraction(abstr_args, body),
        terms=output_vars,
        meta=None
    )
    return reduce

def _translate_to_formula(ctx: TranslationCtx, task: ir.Task) -> lqp.Formula:
    if isinstance(task, ir.Logical):
        conjuncts = [_translate_to_formula(ctx, child) for child in task.body]
        return mk_and(conjuncts)
    elif isinstance(task, ir.Lookup):
        return _translate_to_atom(ctx, task)
    elif isinstance(task, ir.Not):
        return lqp.Not(arg=_translate_to_formula(ctx, task.task), meta=None)
    elif isinstance(task, ir.Exists):
        lqp_vars, conjuncts = _translate_bindings(ctx, list(task.vars))
        conjuncts.append(_translate_to_formula(ctx, task.task))
        return mk_exists(lqp_vars, mk_and(conjuncts))
    elif isinstance(task, ir.Construct):
        assert len(task.values) >= 1, "Construct should have at least one value"
        terms = [_translate_term(ctx, arg) for arg in task.values]
        result_term = _translate_term(ctx, task.id_var)
        terms.append(result_term)
        assert result_term[1].type_name == lqp.TypeName.UINT128, \
            f"Attempting to store a {task.id_var} in a type `{result_term[1].type_name}`"

        return mk_primitive("rel_primitive_hash_tuple_uint128", [v for v, _ in terms])
    elif isinstance(task, ir.Union):
        # TODO: handle hoisted vars if needed
        disjs = [_translate_to_formula(ctx, child) for child in task.tasks]
        return mk_or(disjs)
    elif isinstance(task, (ir.Aggregate, ir.Output, ir.Update)):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    elif isinstance(task, ir.Rank):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    else:
        raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

# Only used for translating terms on relatoms, which can be specialized values.
def _translate_relterm(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.RelTerm, lqp.Type]:
    if isinstance(term, ir.Literal) and term.type == types.Symbol:
        if isinstance(term.value, str):
            value = mk_value(term.value)
            return mk_specialized_value(value), meta_type_to_lqp(types.String)
        elif isinstance(term.value, int):
            value = mk_value(term.value)
            return mk_specialized_value(value), meta_type_to_lqp(types.Int64)
        else:
            raise NotImplementedError(f"Cannot specialize literal of type {type(term.value)}")
    return _translate_term(ctx, term)

def _translate_term(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.Term, lqp.Type]:
    if isinstance(term, ir.ScalarType):
        # TODO: ScalarType is not like other terms, should be handled separately.
        return to_lqp_value(term.name, types.String), meta_type_to_lqp(types.String)
    elif isinstance(term, ir.Var):
        t = meta_type_to_lqp(term.type)
        return _translate_var(ctx, term), t
    else:
        assert isinstance(term, ir.Literal), f"Cannot translate value {term!r} of type {type(term)} to LQP Term; neither Var nor Literal."
        v = to_lqp_value(term.value, term.type)
        return v, meta_type_to_lqp(term.type)

def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    if task.relation == builtins.cast:
        assert len(task.args) == 3, f"expected three terms for {task.relation.name}, got {len(task.args)}"

        terms = []
        for arg in task.args[1:]:
            term, _ = _translate_relterm(ctx, arg)
            terms.append(term)
        return lqp.Cast(input=terms[0], result=terms[1], meta=None)

    if task.relation == builtins.join:
        return _translate_join(ctx, task)
    elif task.relation == builtins.infomap:
        return _translate_infomap(ctx, task)
    elif task.relation == builtins.louvain:
        return _translate_louvain(ctx, task)
    elif task.relation == builtins.label_propagation:
        return _translate_label_propagation(ctx, task)

    terms = []
    term_types = []
    for arg in task.args:
        # Handle varargs, which come wrapped in a tuple.
        if isinstance(arg, tuple):
            for vararg in arg:
                term, ty = _translate_relterm(ctx, vararg)
                terms.append(term)
                term_types.append(ty)
        else:
            term, ty = _translate_relterm(ctx, arg)
            terms.append(term)
            term_types.append(ty)

    if builtins.is_pragma(task.relation):
        lqp_name = pragma_to_lqp_name(task.relation.name)
        return mk_pragma(lqp_name, terms)

    if builtins.is_builtin(task.relation):
        return build_primitive(task.relation.name, terms, term_types)

    if helpers.is_external(task.relation):
        return lqp.RelAtom(name=task.relation.name, terms=terms, meta=None)

    projection, _ = _translate_bindings(ctx, list(task.args))
    rid = get_relation_id(ctx, task.relation, projection)
    return lqp.Atom(name=rid, terms=terms, meta=None)


def get_relation_id(ctx: TranslationCtx, relation: ir.Relation, projection: list[Tuple[lqp.Var, lqp.Type]] = []) -> lqp.RelationId:
    types = "_".join([str(t.type_name) for (_, t) in projection])
    if types:
        types = "_" + types
    if relation.id in ctx.def_names.id_to_name:
        unique_name = ctx.def_names.id_to_name[relation.id]
    else:
        name = helpers.relation_name_prefix(relation) + relation.name
        name = helpers.sanitize(name)
        unique_name = ctx.def_names.get_name_by_id(relation.id, name)

    return utils.gen_rel_id(ctx, unique_name, types)

def get_output_id(ctx: TranslationCtx, orig_name: str, metamodel_id: int) -> lqp.RelationId:
    unique_name = ctx.output_names.get_name_by_id(metamodel_id, orig_name)
    return utils.gen_rel_id(ctx, unique_name)

def _translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.Type]], list[lqp.Formula]]:
    lqp_vars = []
    conjuncts = []
    for binding in bindings:
        lqp_var, typ, eq = binding_to_lqp_var(ctx, binding)
        lqp_vars.append((lqp_var, typ))
        if eq is not None:
            conjuncts.append(eq)

    return lqp_vars, conjuncts

def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp.Var, lqp.Type, Union[None, lqp.Formula]]:
    if isinstance(binding, ir.Var):
        var, typ = _translate_term(ctx, binding)
        assert isinstance(var, lqp.Var)
        return var, typ, None
    elif isinstance(binding, ir.Literal):
        lqp_value = to_lqp_value(binding.value, binding.type)
        var, formula = constant_to_var(ctx, lqp_value)
        return var, meta_type_to_lqp(binding.type), formula
    else:
        raise Exception(f"Unsupported binding type: {type(binding)}")

def to_lqp_value(value: ir.PyValue, value_type: ir.Type) -> lqp.Value:
    typ = meta_type_to_lqp(value_type)

    # Ensure int values match the requested integer type.
    if typ.type_name == lqp.TypeName.INT and isinstance(value, int):
        assert lqp_types.INT_MIN <= value <= lqp_types.INT_MAX, f"{value} out of range for a 64-bit INT value"
        val = value
    elif typ.type_name == lqp.TypeName.INT128 and isinstance(value, int):
        assert lqp_types.INT128_MIN <= value <= lqp_types.INT128_MAX, f"{value} out of range for an INT128 value"
        val = lqp.Int128Value(value=value, meta=None)
    elif typ.type_name == lqp.TypeName.UINT128 and isinstance(value, int):
        assert lqp_types.UINT128_MIN <= value <= lqp_types.UINT128_MAX, f"{value} out of range for an UINT128 value"
        val = lqp.UInt128Value(value=value, meta=None)
    elif typ.type_name == lqp.TypeName.FLOAT and isinstance(value, float):
        val = value
    elif typ.type_name == lqp.TypeName.STRING and isinstance(value, str):
        val = value
    elif typ.type_name == lqp.TypeName.DECIMAL and isinstance(value, (int, float, PyDecimal)):
        precision = typ.parameters[0].value
        scale = typ.parameters[1].value
        assert isinstance(precision, int) and isinstance(scale, int)
        val = lqp.DecimalValue(precision=precision, scale=scale, value=PyDecimal(value), meta=None)
    elif typ.type_name == lqp.TypeName.DATE and isinstance(value, date):
        val = lqp.DateValue(value=value, meta=None)
    elif typ.type_name == lqp.TypeName.DATETIME and isinstance(value, datetime):
        utc_value = value.astimezone(timezone.utc) if value.tzinfo is not None else value # Convert to UTC cf. Iceberg
        val = lqp.DateTimeValue(value=utc_value, meta=None)
    elif typ.type_name == lqp.TypeName.BOOLEAN and isinstance(value, bool):
        val = lqp.BooleanValue(value=value, meta=None)
    else:
        raise Exception(f"Unsupported type for constant: {typ.type_name} with value {value} of type {type(value)}")

    return mk_value(val)

def constant_to_var(ctx: TranslationCtx, value: lqp.Value, name_hint: str = "cvar") -> Tuple[lqp.Var, lqp.Formula]:
    var = gen_unique_var(ctx, name_hint)
    eq = mk_primitive("rel_primitive_eq", [var, value])
    return var, eq

def _extract_pyrel_error_ids(ctx: TranslationCtx, model: ir.Model) -> list[Tuple[lqp.RelationId, str]]:
    effects = collect_by_type(ir.Update, model)
    pyrel_error_attrs = []
    # We use a separate counter here to avoid modifying the current model.
    i = 1
    for effect in effects:
        if "pyrel_error_attrs" not in effect.relation.name:
            continue
        projection, _ = _translate_bindings(ctx, _effect_bindings(effect))
        rel_id = get_relation_id(ctx, effect.relation, projection)
        name = f"{effect.relation.name}_{i}"
        pyrel_error_attrs.append((rel_id, name))
        i += 1

    return pyrel_error_attrs

# Translate a relation reference into an abstraction over its fields.
def _translate_relation_ref(ctx: TranslationCtx, relation: ir.Relation) -> lqp.Abstraction:
    projection = []
    for field in relation.fields:
        var = gen_unique_var(ctx, field.name)
        typ = meta_type_to_lqp(field.type)
        projection.append((var, typ))

    rid = get_relation_id(ctx, relation, projection)
    atom = lqp.Atom(name=rid, terms=[var for (var, _) in projection], meta=None)
    return mk_abstraction(projection, atom)

# Common translation logic for graph algorithms.
# task.args[0] : normalized weight list (int64, int64, float)
# task.args[1] : normalized node count (relation or constant int64)
# task.args[2] : normalized edge count (relation or constant int64)
# task.args[3:-3] : algorithm parameters (var or constant)
# task.args[-3] : diagnostic info
# task.args[-2] : node index
# task.args[-1] : community ident
def _translate_graph_common(name: str, ctx: TranslationCtx, task: ir.Lookup):
    abstractions = []

    assert isinstance(task.args[0], ir.Relation), \
        f"Expected relation as first arg to {name}, got {task.args[0]}:{type(task.args[0])}"
    abstractions.append(_translate_relation_ref(ctx, task.args[0]))

    # Allow constant args for node and edge count
    for arg in task.args[1:3]:
        if isinstance(arg, ir.Relation):
            abst = _translate_relation_ref(ctx, arg)
            typ = abst.vars[0][1]
            assert typ.type_name == lqp.TypeName.INT, \
                f"Expected Int64 types for node and edge counts, got type {typ.type_name}"
            abstractions.append(abst)
        else:
            var, typ, eq = binding_to_lqp_var(ctx, arg)
            assert eq is not None, \
                f"Expected equality formula for {name} arg {arg}:{type(arg)}"
            abstractions.append(mk_abstraction([(var, typ)], mk_and([eq])))

    for arg in task.args[3:-3]:
        var, typ, eq = binding_to_lqp_var(ctx, arg)
        if eq:
            abstractions.append(mk_abstraction([(var, typ)], mk_and([eq])))
        else:
            print(f"Primitive graph algorithm arg without eq:\n var:{var}, typ:{typ}\n arg:{arg}:{type(arg)}")
            abstractions.append(mk_abstraction([(var, typ)], mk_and([])))

    terms = []
    for arg in task.args[-3:]:
        term, _ = _translate_relterm(ctx, arg)
        terms.append(term)

    return (abstractions, terms)

def _translate_infomap(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    abstractions, terms = _translate_graph_common("infomap", ctx, task)

    return lqp.FFI(
        meta=None,
        name="rel_primitive_infomap",
        args=abstractions,
        terms=terms,
    )

def _translate_louvain(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    abstractions, terms = _translate_graph_common("louvain", ctx, task)

    return lqp.FFI(
        meta=None,
        name="rel_primitive_louvain",
        args=abstractions,
        terms=terms,
    )

def _translate_label_propagation(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    abstractions, terms = _translate_graph_common("label_propagation", ctx, task)

    return lqp.FFI(
        meta=None,
        name="rel_primitive_async_label_propagation",
        args=abstractions,
        terms=terms,
    )

# Hard-coded implementation of Rel's string_join
def _translate_join(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
    assert len(task.args) == 3
    (strs, separator, target) = task.args
    assert isinstance(separator, ir.Node)
    assert isinstance(target, ir.Var)
    assert isinstance(strs, tuple)

    # Enumerate the strings we're going to be reducing over
    enumerated_conjunctions = []
    index_var, index_type = gen_unique_var(ctx, "idx"), meta_type_to_lqp(types.Int64)
    string_var, string_type = gen_unique_var(ctx, "s"), meta_type_to_lqp(types.String)
    for i, s in enumerate(strs):
        index = to_lqp_value(i, types.Int64)
        string = _translate_term(ctx, s)[0]
        eq_idx = mk_primitive("rel_primitive_eq", [index_var, index])
        eq_string = mk_primitive("rel_primitive_eq", [string_var, string])
        enumerated_conjunctions.append(mk_and([eq_idx, eq_string]))
    body = mk_abstraction([(index_var, index_type), (string_var, string_type)], mk_or(enumerated_conjunctions))

    # Make the function we're reducing
    a_var, a_type = gen_unique_var(ctx, "a"), meta_type_to_lqp(types.String)
    b_var, b_type = gen_unique_var(ctx, "b"), meta_type_to_lqp(types.String)
    curr_var, curr_type = gen_unique_var(ctx, "curr"), meta_type_to_lqp(types.String)
    res_var, res_type = gen_unique_var(ctx, "res"), meta_type_to_lqp(types.String)

    sep, _ = _translate_term(ctx, separator)

    op = mk_abstraction(
        [(a_var, a_type), (b_var, b_type), (curr_var, curr_type)],
        mk_exists([(res_var, res_type)],
            mk_and([
                mk_primitive("rel_primitive_concat", [res_var, b_var, curr_var]),
                mk_primitive("rel_primitive_concat", [a_var, sep, res_var])
            ])
        )
    )

    output_term = _translate_term(ctx, target)[0]

    return lqp.Reduce(meta=None, op=op, body=body, terms=[output_term])

def _translate_var(ctx: TranslationCtx, term: ir.Var) -> lqp.Var:
    name = ctx.var_names.get_name_by_id(term.id, term.name)
    return lqp.Var(name=name, meta=None)



"""Base frontend classes for RelationalAI semantics API.
"""
from __future__ import annotations

import decimal
import math
import re
import datetime as dt
from typing import TYPE_CHECKING, Any, Generic, Iterable, Iterator, NoReturn, Optional, Sequence, Tuple, Type, TypeGuard, TypeVar, cast
import itertools
from more_itertools import peekable
from pandas import DataFrame
import pandas as pd
from enum import Enum, EnumMeta

from ...util import schema as schema_util
from ...util.structures import KeyedSet
from ...util.naming import Namer, sanitize
from ...util.python import pytype_to_concept_name
from ...util.source import SourcePos
from ...util.tracing import get_tracer
from ...util.error import err, exc, warn, source
from ...util.docutils import include_in_docs
from ..metamodel.builtins import builtins
from ..metamodel.metamodel import Model as mModel

tracer = get_tracer()
tracer.start_program()

#------------------------------------------------------
# Global ID Generator
#------------------------------------------------------

_global_id = peekable(itertools.count(0))

#------------------------------------------------------
# Helpers
#------------------------------------------------------

def _find_field(fields: list[Field], field: str|int|Concept) -> tuple[Field, int]|None:
    resolved = None
    if isinstance(field, int):
        if not (0 <= field < len(fields)):
            raise IndexError(f"Field index {field} out of range for relationship with {len(fields)} fields")
        resolved = fields[field]
    elif isinstance(field, str):
        resolved = next((f for f in fields if f.name == field), None)
    elif isinstance(field, Concept):
        resolved = next((f for f in fields if f.type is field), None)
    if not resolved:
        return None
    return resolved, fields.index(resolved)

def _find_concept(model: Model, passed_type: Any, extra_types: Sequence[Concept] = [], default:Concept|None = None) -> Concept | None:
    if isinstance(passed_type, Concept):
        return passed_type
    if isinstance(passed_type, str):
        for ext_type in extra_types:
            if ext_type._name == passed_type:
                return ext_type
        name = pytype_to_concept_name.get(passed_type, passed_type)
        if name.startswith("Number(") or name.startswith("Decimal("):
            return cast(Any, CoreConcepts["Number"]).parse(name)
        return model._find_concept(name)
    #TODO: try to find a variable in the frame that is a concept with this name
    return default

#------------------------------------------------------
# Primitive
#------------------------------------------------------

Primitive = str|int|float|bool

def is_primitive(value: Any) -> TypeGuard[str | int | float | bool]:
    return isinstance(value, (str, int, float, bool))

#------------------------------------------------------
# DSLBase
#------------------------------------------------------

class DSLBase:
    def __init__(self, model:Model):
        self._model = model
        self._id = next(_global_id)
        self._source = SourcePos.new()

    #------------------------------------------------------
    # Hashing
    #------------------------------------------------------

    def to_dict_key(self) -> int|tuple[int, int]:
        return self._id

    def __hash__(self) -> NoReturn:
        exc("Unhashable type", f"{self.__class__.__name__} objects are unhashable. Use `to_dict_key()` instead.", [source(self)])

#------------------------------------------------------
# DSLDict
#------------------------------------------------------

def dsl_key(obj:Any) -> Any:
    if isinstance(obj, DSLBase):
        return obj.to_dict_key()
    elif isinstance(obj, tuple):
        return tuple(dsl_key(o) for o in obj)
    return obj

#------------------------------------------------------
# Variable
#------------------------------------------------------

class Variable(DSLBase):
    #--------------------------------------------------
    # Infix operator overloads
    #--------------------------------------------------

    def _bin_op(self, op, left, right) -> Expression:
        res = CoreConcepts["Numeric"].ref()
        return Expression(CoreRelationships[op], [left, right, res])

    def __add__(self, other):
        return self._bin_op("+", self, other)
    def __radd__(self, other):
        return self._bin_op("+", other, self)

    def __mul__(self, other):
        return self._bin_op("*", self, other)
    def __rmul__(self, other):
        return self._bin_op("*", other, self)

    def __sub__(self, other):
        return self._bin_op("-", self, other)
    def __rsub__(self, other):
        return self._bin_op("-", other, self)

    def __truediv__(self, other):
        return self._bin_op("/", self, other)
    def __rtruediv__(self, other):
        return self._bin_op("/", other, self)

    def __floordiv__(self, other):
        return self._bin_op("//", self, other)
    def __rfloordiv__(self, other):
        return self._bin_op("//", other, self)

    def __pow__(self, other):
        return self._bin_op("^", self, other)
    def __rpow__(self, other):
        return self._bin_op("^", other, self)

    def __mod__(self, other):
        return self._bin_op("%", self, other)
    def __rmod__(self, other):
        return self._bin_op("%", other, self)

    def __neg__(self):
        return self._bin_op("*", self, -1)

    #--------------------------------------------------
    # Filter overloads
    #--------------------------------------------------

    def _filter(self, op, left, right) -> Expression:
        return Expression(CoreRelationships[op], [left, right])

    def __gt__(self, other):
        return self._filter(">", self, other)
    def __ge__(self, other):
        return self._filter(">=", self, other)
    def __lt__(self, other):
        return self._filter("<", self, other)
    def __le__(self, other):
        return self._filter("<=", self, other)
    def __eq__(self, other) -> Any:
        return self._filter("=", self, other)
    def __ne__(self, other) -> Any:
        return self._filter("!=", self, other)

    #--------------------------------------------------
    # And/Or
    #--------------------------------------------------

    def __or__(self, other) -> Match:
        return Match(self._model, self, other)

    def __and__(self, other) -> Fragment:
        if isinstance(other, Fragment):
            new = other.where()
            new._where.insert(0, self)
            return new
        return self._model.where(self, other)

    #------------------------------------------------------
    # AsBool
    #------------------------------------------------------

    def as_bool(self) -> AsBool:
        return AsBool(self)

    #------------------------------------------------------
    # Alias
    #------------------------------------------------------

    def alias(self, alias: str) -> Alias:
        return Alias(self, alias)

    #--------------------------------------------------
    # in_
    #--------------------------------------------------

    def in_(self, values:Sequence[Value]|Variable):
        if isinstance(values, Variable):
            return self == values
        if all(isinstance(v, (str, int, float, bool)) for v in values):
            data_table = self._model.data([(v,) for v in values])
            return self == data_table[0]
        else:
            return self == self._model.union(*values)

    #--------------------------------------------------
    # Check value
    #--------------------------------------------------

    def _check_value(self) -> bool:
        return True

    def _to_concept(self) -> Concept|None:
        return None

    #------------------------------------------------------
    # Error handling
    #------------------------------------------------------

    def __bool__(self) -> NoReturn:
        cur_source = self._source.block.source or ""
        invalid = next((bool_check for bool_check in ["if ", "while ", " and ", " or "] if bool_check in cur_source), "bool check").strip()
        mapped = {"and": "`&` or `,`", "or": "`|`", "if": "`where`"}
        if m := mapped.get(invalid):
            exc(f"Invalid operator", f"Cannot use python's `{invalid}` in model expressions. Use {m} instead.", [source(self)])
        else:
            exc(f"Invalid operator", f"Cannot use python's `{invalid}` in model expressions.", [source(self)])

    if not TYPE_CHECKING:
        def __iter__(self):
            common_incorrect = ["sum", "min", "max"]
            for agg in common_incorrect:
                if agg in self._source.block.source:
                    exc(f"Invalid built-in", f"The Python built-in `{agg}()` was used instead of the RAI equivalent.", [
                        source(self),
                        f"Use [cyan]`relationalai.semantics.std.aggregates.{agg}`[/cyan] instead.",
                    ])
            exc("Invalid iteration", f"Cannot iterate over {self.__class__.__name__} objects.", [
                source(self)
            ])

        def __len__(self) -> NoReturn:
            exc("Invalid operator", "Cannot use len() on model variables.", [source(self)])

#------------------------------------------------------
# Concept
#------------------------------------------------------

class Concept(Variable):
    lookup_by_id: dict[int, Concept] = {}

    def __init__(self, name: str, extends: list[Concept], identify_by: dict[str, Concept], model:Model):
        super().__init__(model)
        Concept.lookup_by_id[self._id] = self
        self._name = name
        clean_extends = []
        for extend in extends:
            if not isinstance(extend, Concept):
                exc("Invalid extend", f"Cannot extend from non-Concept type: {extend}", [source(self)])
            if "Number" in CoreConcepts and extend is CoreConcepts["Number"]:
                clean_extends.append(extend.size(38, 14))
            else:
                clean_extends.append(extend)
        self._extends = clean_extends
        self._annotations: list[Expression|Relationship] = []
        self._relationships = {}
        self._identify_by:list[Relationship] = []
        for id_name, id_val in identify_by.items():
            if not isinstance(id_val, (Concept, Property)):
                exc("Invalid identify_by", f"identify_by values must be Concepts, got: {type(id_val)}", [source(self)])
            id_val = self._dot(id_name, with_type=id_val)
            self._identify_by.append(id_val)

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, Relationship):
            key = name.lower()
            if self._relationships.get(key) is not None:
                exc("Duplicate relationship", f"Relationship '{name}' is already defined on concept '{self._name}'", [
                    source(self._relationships[key])
                ])
            if not value._short_name:
                value._short_name = name
            self._relationships[key] = value
        else:
            super().__setattr__(name, value)

    def _dot_recur(self, name: str, with_type: Any = None) -> Relationship|None:
        if name.lower() in self._relationships:
            return self._relationships[name.lower()]
        for ext in self._extends:
            rel = ext._dot_recur(name, with_type)
            if rel is not None:
                return rel
        # We should create the relationship on the root type if it's not
        # found anywhere
        if self._model is not CoreLibrary:
            field_type = _find_concept(self._model, with_type, default=CoreConcepts["Any"])
            assert field_type is not None
            rel = self._relationships[name.lower()] = Property(fields=[
                Field(self._name.lower(), self),
                Field(name, field_type)
            ], short_name=name, model=self._model)
            return rel
        return None

    def _dot(self, name: str, with_type: Any = None) -> Relationship:
        if self._model is CoreLibrary:
            exc("Invalid relationship", f"Cannot access relationships on core concept '{self._name}'.", [source(self)])
        if with_type is None:
            with_type = CoreConcepts["Any"]
        rel = self._dot_recur(name, with_type)
        assert rel is not None
        return rel

    def __getattr__(self, item) -> Chain:
        if item.startswith("_"):
            return object.__getattribute__(self, item)
        return Chain(self, self._dot(item))

    def __call__(self, *args: Any, **kwargs: Any) -> Expression:
        return Expression(self, list(args), kwargs)

    def new(self, *args: StatementAndSchema, **kwargs: Any) -> New:
        return New(self, args, kwargs)

    def to_identity(self, *args: Any, **kwargs: Any) -> New:
        return New(self, args, kwargs, identity_only=True)

    def identify_by(self, *properties: Property|Chain) -> Concept:
        for prop in properties:
            if isinstance(prop, Chain):
                prop = prop._next
            if not isinstance(prop, Property) and not (isinstance(prop, Reading) and isinstance(prop._relationship, Property)):
                exc("Invalid identify_by", f"identify_by expects Properties, got: {type(prop).__name__}", [source(self)])
            if prop._fields[0].type is not self:
                exc("Invalid identify_by", f"Property {prop} have the first field be {self._name}", [source(self)])
            self._identify_by.append(prop)
        return self

    def filter_by(self, **kwargs: Any) -> Expression:
        return FilterBy(self, kwargs)

    def require(self, *items: Variable|Fragment) -> Fragment:
        return self._model.where(self).require(*items)

    def ref(self, name="") -> Ref:
        return Ref(self, name)

    def annotate(self, *annos:Expression|Relationship) -> Concept:
        self._annotations.extend(annos)
        return self

    def _to_concept(self) -> Concept:
        return self

    def __dir__(self):
        default = set(super().__dir__())
        return sorted(default.union(self._relationships.keys()))

    def __format__(self, format_spec: str) -> str:
        if not format_spec:
            return f"{{{self._name}#{self._id}}}"
        return f"{{{self._name}#{self._id}:{format_spec.strip()}}}"

#------------------------------------------------------
# NumberConcept
#------------------------------------------------------

class NumberConcept(Concept):
    def __init__(self, name: str, precision: int, scale: int, model: Model):
        super().__init__(name, extends=[CoreConcepts["Numeric"]], identify_by={}, model=model)
        self._precision = precision
        self._scale = scale

#------------------------------------------------------
# Ref
#------------------------------------------------------

class Ref(Variable):
    def __init__(self, concept: Concept, name: str|None = None):
        super().__init__(concept._model)
        self._concept = concept
        self._name = name or ("number" if (isinstance(concept, NumberConcept) or concept is CoreConcepts["Numeric"])
                              else concept._name)

    def __getattr__(self, item):
        if item.startswith("_"):
            return object.__getattribute__(self, item)
        return Chain(self, self._concept._dot(item))

    def _to_concept(self) -> Concept:
        return self._concept

    def FilterBy(self, **kwargs: Any) -> Expression:
        return FilterBy(self, kwargs)

#------------------------------------------------------
# Table
#------------------------------------------------------

class Table(Concept):
    def __init__(self, name:str, schema: dict[str, Concept], model: Model):
        super().__init__(name, [], {}, model)
        self._known_columns = []
        for col_name, col_type in schema.items():
            if isinstance(col_type, Property):
                col_rel = col_type
                setattr(self, col_name, col_rel)
            else:
                col_rel = self._dot(col_name, with_type=col_type)
            self._known_columns.append(col_rel)

    @property
    def _columns(self) -> list[Relationship]:
        if self._known_columns:
            return self._known_columns
        schema = schema_util.fetch(self._name)
        for col_name, col_type_name in schema.items():
            col_type = _find_concept(self._model, col_type_name, default=CoreConcepts["Any"])
            col_rel = self._dot(col_name, with_type=col_type)
            col_rel._short_name = col_name
            if col_type is not None:
                col_rel._fields[1].type = col_type
            self._known_columns.append(col_rel)
        return self._known_columns

    def __getitem__(self, index: int|str) -> Chain:
        if isinstance(index, int):
            if not (0 <= index < len(self._columns)):
                raise IndexError(f"Column index {index} out of range, there are {len(self._columns)} columns")
            return Chain(self, self._columns[index])
        col = next((c for c in self._columns if c._short_name == index), None)
        if col is None:
            raise KeyError(f"Column name '{index}' not found, columns have names {[c._short_name for c in self._columns]}")
        return Chain(self, col)

    def __iter__(self) -> Iterator[Relationship]:
        return iter(self._columns)

    def new(self, *args:StatementAndSchema, **kwargs: Any) -> NoReturn:
        exc("Invalid new call", "Cannot create new instances of Tables.", [source(self)])

    def to_identity(self, *args: Any, **kwargs: Any) -> NoReturn:
        exc("Invalid identity call", "Cannot create identity instances of Tables.", [source(self)])

    def to_schema(self, *, exclude: list[str] = []) -> TableSchema:
        return TableSchema(self, exclude=exclude)

#------------------------------------------------------
# TableSchema
#------------------------------------------------------

class TableSchema(DSLBase):
    def __init__(self, table: Table, exclude: list[str] = []):
        super().__init__(table._model)
        self._table = table
        self.exclude = set([e.lower() for e in exclude])

    def get_columns(self) -> list[Relationship]:
        return [col for col in self._table._columns if col._short_name.lower() not in self.exclude]

#--------------------------------------------------
# DerivedTable
#--------------------------------------------------

class DerivedTable(Variable):
    def __init__(self, model: Model):
        super().__init__(model)
        self.__cols = []

    @property
    def _columns(self):
        if not self.__cols:
            self.__cols = self._get_cols()
        if not self.__cols:
            raise ValueError(f"Cannot use {self.__class__.__name__} as it has no columns")
        return self.__cols

    def _check_value(self):
        return bool(self._columns)

    def __iter__(self) -> Iterator[DerivedColumn]:
        return iter(self._columns)

    def __getitem__(self, index: int|str) -> DerivedColumn:
        if isinstance(index, int):
            if not (0 <= index < len(self._columns)):
                raise IndexError(f"Column index {index} out of range, there are {len(self._columns)} columns")
            return self._columns[index]
        col = next((c for c in self._columns if c._name == index), None)
        if not col:
            raise KeyError(f"Column name '{index}' not found, columns have names {[c._name for c in self._columns]}")
        return col

    def _get_cols(self):
        raise NotImplementedError()

#--------------------------------------------------
# DerivedColumn
#--------------------------------------------------

class DerivedColumn(Variable):
    def __init__(self, table: DerivedTable, index: int, name: str|None = None, type_: Concept|None = None):
        super().__init__(table._model)
        self._table = table
        self._index = index
        self._name = name
        self._type = type_
        self._relationships = {}

    def _dot(self, name: str) -> Relationship:
        if self._type is not None:
            return self._type._dot(name)
        if name.lower() not in self._relationships:
            self._relationships[name.lower()] = Property(fields=[
                Field("entity", CoreConcepts["Any"]),
                Field(name, CoreConcepts["Any"])
            ], short_name=name, model=self._model)
        return self._relationships[name.lower()]

    def __getattr__(self, item):
        if item.startswith("_"):
            return object.__getattribute__(self, item)
        return Chain(self, self._dot(item))

    @staticmethod
    def from_value(table: DerivedTable, index: int, var: Value) -> DerivedColumn:
        name = None
        if isinstance(var, Alias):
            name = var._alias
        if not isinstance(var, Variable):
            var = Literal(var, table._model)
        return DerivedColumn(table=table, index=index, name=name, type_=var._to_concept())

#------------------------------------------------------
# Field
#------------------------------------------------------

class Field:
    def __init__(self, name: str, type_: Concept, is_input: bool = False, is_list: bool = False, source: SourcePos | None = None):
        self._id = next(_global_id)
        self.name = name
        self.type = type_
        self.is_input = is_input
        self.is_list = is_list
        self._source = source

    @classmethod
    def input(cls, name: str, concept: Concept, is_list: bool = False) -> Field:
        return Field(name, concept, is_input=True, is_list=is_list)

    def _match(self, other: Field) -> bool:
        return self.name == other.name and self.type is other.type

#--------------------------------------------------
# Relationship
#--------------------------------------------------

class Relationship(Variable):
    def __init__(self, model: Model, reading_str:str = "", fields: list[Field] = [], short_name: str = "", allow_no_fields: bool = False, overloads:list[list[Concept]]|None = None, is_unresolved: bool = False):
        super().__init__(model)
        if not reading_str and not fields and not allow_no_fields:
            raise ValueError("Either reading_str or fields must be provided")
        if not reading_str and fields:
            reading_str = " and ".join([f"{{{f.type._name}#{f.type._id}:{f.name}}}" for f in fields[:-1]])
            reading_str = f"{reading_str} has {{{fields[-1].type._name}#{fields[-1].type._id}:{fields[-1].name}}}"
        parts = []
        if not fields:
            (fields, parts) = Reading.parse(model, reading_str, fields, _source=self)
        self._fields = fields
        if not fields and not allow_no_fields:
            exc("Invalid Relationship", "A Relationship must have at least one field.", [source(self)])
        self._readings = [Reading(model, self, reading_str, fields, parts)]
        self._short_name = short_name
        self._relationships = {}
        self._annotations: list[Expression|Relationship] = []
        self._overloads = overloads
        self._is_unresolved = is_unresolved
        for f in self._fields:
            if f._source is None:
                f._source = self._source

    def _dot(self, name: str) -> Relationship:
        field_type = self._fields[-1].type
        if name.lower() in self._relationships:
            return self._relationships[name.lower()]
        if field_type is CoreConcepts["Any"]:
            rel = Property(fields=[
                Field(self._fields[-1].name, field_type),
                Field(name, CoreConcepts["Any"])
            ], short_name=name, model=self._model, is_unresolved=True)
            self._relationships[name.lower()] = rel
            return rel
        return field_type._dot(name)

    def __getattr__(self, item):
        if item.startswith("_"):
            return object.__getattribute__(self, item)
        op = self._fields[-1].type._dot(item)
        return Chain(self, op)

    def __call__(self, *args: Any, **kwargs: Any) -> Expression:
        return Expression(self, list(args), kwargs, root=self)

    def __getitem__(self, field: str|int|Concept) -> FieldRef:
        resolved = _find_field(self._fields, field)
        if not resolved:
            raise KeyError(f"Field {field} not found in relationship with fields {[f.name for f in self._fields]}")
        return FieldRef(self, field, *resolved)

    def alt(self, reading_str: str) -> Reading:
        reading = Reading(self._model, self, reading_str)
        self._readings.append(reading)
        return reading

    def annotate(self, *annos:Expression|Relationship) -> Relationship:
        self._annotations.extend(annos)
        return self

    def _to_concept(self) -> Concept|None:
        return self._fields[-1].type

    def __dir__(self):
        default = set(super().__dir__())
        return sorted(default.union(self._fields[-1].type.__dir__()))

    def to_df(self) -> DataFrame:
        refs = [self[ix] for ix in range(len(self._fields))]
        return self._model.select(*refs).where(self(*refs)).to_df()

    def inspect(self):
        print(self.to_df())

#------------------------------------------------------
# Reading
#------------------------------------------------------

class Reading(Relationship):
    def __init__(self, model: Model, relationship: Relationship, reading_str: str, reading_fields: list[Field] = [], reading_parts: list[str|int] = []):
        Variable.__init__(self, model)
        self._relationship = relationship
        if not reading_fields or not reading_parts:
            parsed_fields, parsed_parts = Reading.parse(model, reading_str, reading_fields, _source=self)
            reading_fields = reading_fields or parsed_fields
            reading_parts = reading_parts or parsed_parts
        # make sure we can find all these fields in the base relationship
        index_map = {}
        matched = []
        for field_i, field in enumerate(reading_fields):
            found = next(((i, f) for i, f in enumerate(relationship._fields) if f._match(field)), None)
            if not found:
                raise ValueError(f"Field {field.name}:{field.type._name} not found in relationship with fields {[f'{f.name}:{f.type._name}' for f in relationship._fields]}")
            matched.append(found[1])
            index_map[field_i] = found[0]
        self._fields = matched
        self._reading = reading_str
        self._short_name = ""
        self._parts:list[str|int] = [index_map.get(p, p) for p in reading_parts]
        self._relationships = {}
        self._annotations: list[Expression|Relationship] = []

    #------------------------------------------------------
    # Parse
    #------------------------------------------------------

    @classmethod
    def parse(cls, model: Model, reading: str, known_fields: list[Field], _source=None) -> Tuple[list[Field], list[str|int]]:
        # match <class 'foo'> which is the serialized form of a python type mistakenly passed instead of a concept
        class_pattern = re.compile(r'<class \'(.*?)\'>')
        match = class_pattern.search(reading)
        if match:
            extra: list = [source(_source)]
            if match.group(1) in pytype_to_concept_name:
                concept = pytype_to_concept_name[match.group(1)]
                extra.append(f" Did you mean to use [cyan]relationalai.semantics.{concept}[/cyan]?")
            exc("Invalid field type", f"The type '{match.group(1)}' is not a valid Concept.", extra)

        # {Type} or {name:Type}, where Type can include Number(38,14)
        pattern = re.compile(r'\{([a-zA-Z0-9_.#]+(?:(?:\([0-9]+,[0-9]+\))(?:[#0-9]+)?)?)(?::\s*([a-zA-Z0-9_.]+(?:\([0-9]+,[0-9]+\))?))?\}')

        namer = Namer()
        fields: list[Field] = []
        parts: list[str|int] = []

        last_end = 0
        is_old_style = True
        for m in pattern.finditer(reading):
            # literal chunk before this match
            parts.append(reading[last_end:m.start()])
            field_name, field_type_name = m.group(1), m.group(2)
            field_type_id = None
            if "#" in field_name:
                temp_name = field_type_name
                field_type_name, field_type_id = field_name.split("#")
                field_name = temp_name or sanitize(field_type_name.lower())
                is_old_style = False

            # if we don't have a type_name, then only a type was provided
            if not field_type_name:
                field_type_name = field_name
                field_name = sanitize(field_name.lower())

            field_name = namer.get_name(field_name)
            field_type = Concept.lookup_by_id.get(int(field_type_id)) \
                            if field_type_id \
                            else _find_concept(model, field_type_name, extra_types=[f.type for f in known_fields])
            if field_type is None:
                exc("Unknown Concept", f"The Concept '{field_type_name}' couldn't be found in the model", [
                    source(_source),
                ])

            fields.append(Field(field_name, field_type))
            parts.append(len(fields) - 1)
            last_end = m.end()

        # trailing literal after the final match
        if(last_end < len(reading)):
            parts.append(reading[last_end:])

        if is_old_style and fields:
            correct = []
            for part in parts:
                if isinstance(part, int):
                    type_name = fields[part].type._name
                    correct_type_name = type_name
                    if isinstance(fields[part].type, NumberConcept):
                        precision = fields[part].type._precision
                        scale = fields[part].type._scale
                        if scale == 0:
                            correct_type_name = f"Integer"
                        else:
                            correct_type_name = f"Number.size({precision},{scale})"
                    if type_name.lower() != fields[part].name and correct_type_name.lower() != fields[part].name:
                        correct.append(f"{{{correct_type_name}:{fields[part].name}}}")
                    else:
                        correct.append(f"{{{correct_type_name}}}")
                else:
                    correct.append(part)
            warn("Deprecated format", "Plain strings for Relationships/Properties is deprecated. Use an f-string instead.", [
                source(_source),
                f'For example: [cyan]f"{"".join(correct)}"',
            ])

        return fields, parts

#------------------------------------------------------
# Property
#------------------------------------------------------

class Property(Relationship):
    pass

#------------------------------------------------------
# Literal
#------------------------------------------------------

class Literal(Variable):
    def __init__(self, value: Any, model:Model, type: Concept|None = None):
        super().__init__(model)
        self._value = value
        self._type = type if type is not None else self._get_type(value)

    def _to_concept(self) -> Concept:
        return self._type

    @staticmethod
    def _get_type(value: Any) -> Concept:
        if type(value) is float and (math.isnan(value) or math.isinf(value)):
            return CoreConcepts["Float"]
        if type(value) is float:
            fractional_digits = 0 if math.isnan(value) else min(len(str(value).split(".")[1]), 14)
            return CoreConcepts["Decimal"].size(38, fractional_digits) # type: ignore
        if type(value) is decimal.Decimal:
            str_value = format(value, 'f')
            if '.' in str_value:
                _, fractional_part = str_value.split('.')
                scale = min(len(fractional_part), 14)
            else:
                scale = 14
            return CoreConcepts["Decimal"].size(38, scale) # type: ignore
        if type(value) in pytype_to_concept_name:
            return CoreConcepts[pytype_to_concept_name[type(value)]]
        else:
            raise NotImplementedError(f"Literal type not implemented for value type: {type(value)}")

#------------------------------------------------------
# Chain
#------------------------------------------------------

class Chain(Variable):
    def __init__(self, start: Chain|Concept|Relationship|DerivedColumn|Table|Ref|FieldRef|Expression, next: Relationship, is_ref = False):
        super().__init__(start._model)
        self._start = start
        self._next = next
        self._is_ref = is_ref

    def __getattr__(self, item):
        if item.startswith("_"):
            return object.__getattribute__(self._next, item)
        next_rel = self._next._dot(item)
        return Chain(self, next_rel)

    def __call__(self, *args: Any, **kwargs: Any) -> Expression:
        last = self._next
        assert not isinstance(last, Chain)
        if len(last._fields) > len(args):
            return Expression(last, [self._start, *args], kwargs, root=self)
        return Expression(last, [*args], kwargs, root=self)

    def to_dict_key(self) -> int|tuple:
        if self._is_ref:
            return self._id
        return (self._start.to_dict_key(), self._next.to_dict_key())

    def __getitem__(self, field: str|int|Concept) -> FieldRef:
        resolved = _find_field(self._next._fields, field)
        if not resolved:
            raise KeyError(f"Field {field} not found in relationship with fields {[f.name for f in self._next._fields]}")
        return FieldRef(self, field, *resolved)

    def ref(self) -> Chain:
        return Chain(self._start, self._next, is_ref=True)

    def alt(self, reading_str: str) -> Reading:
        return self._next.alt(reading_str)

    def annotate(self, *annos:Expression|Relationship) -> Relationship:
        return self._next.annotate(*annos)

    def _to_concept(self) -> Concept | None:
        return self._next._to_concept()

#------------------------------------------------------
# Expression
#------------------------------------------------------

class Expression(Variable):
    def __init__(self, op: Relationship|Concept, args: Sequence[Value], kwargs: dict|None = None, root: Chain|Relationship|Concept|None = None):
        super().__init__(op._model)
        self._op = op
        self._has_output = isinstance(op, Concept) or (any(not f.is_input for f in op._fields))
        self._auto_filled = False
        self._root = root

        # clean args
        clean_args = []
        for arg in args:
            if isinstance(arg, (ModelEnum, Field, TupleVariable)):
                pass
            elif isinstance(arg, TableSchema):
                exc("Invalid argument", "Cannot use a schema as an argument to an Expression.", [source(self)])
            elif not isinstance(arg, Variable):
                arg = Literal(arg, self._model)
            elif isinstance(arg, (Match, Fragment)):
                arg._columns
            clean_args.append(arg)
        self._args = clean_args

        self._kwargs = kwargs or {}
        # clean kwargs
        for k, v in self._kwargs.items():
            if not isinstance(v, Variable):
                self._kwargs[k] = Literal(v, self._model)

        if isinstance(op, Relationship):
            op_len = len(op._fields)
            arg_len = len(self._args)
            if op_len - arg_len == 1:
                self._args.append(op._fields[-1].type.ref(name=op._fields[-1].name))
                self._auto_filled = True
            elif op_len != arg_len:
                dir = "Too few" if arg_len < op_len - 1 else "Too many"
                exc(f"{dir} args", f"{op._short_name or 'Relationship'} requires {op_len - 1}-{op_len} arguments but got {arg_len}", [
                    source(self),
                ])

    def __getattr__(self, item):
        if item.startswith("_"):
            return object.__getattribute__(self, item)
        rel = self._op._dot(item)
        return Chain(self, rel)

    def __getitem__(self, field: str|int|Concept) -> FieldRef:
        if not isinstance(self._op, Relationship):
            raise TypeError(f"Cannot index into Expression with non-Relationship: {self._op}")
        resolved = _find_field(self._op._fields, field)
        if not resolved:
            raise KeyError(f"Field {field} not found in relationship with fields {[f.name for f in self._op._fields]}")
        return FieldRef(self, field, *resolved)

    def _to_concept(self) -> Concept | None:
        return self._args[-1]._to_concept() if self._args else self._op._to_concept()

#------------------------------------------------------
# New
#------------------------------------------------------

class New(Expression):
    def __init__(self, concept: Concept, args: Sequence[Any], kwargs: dict[str, Any], identity_only: bool = False):
        clean_args = []
        row_ids = []
        for arg in args:
            if isinstance(arg, TableSchema):
                row_ids.append(arg._table)
                lower_case_kwargs = {k.lower() for k, v in kwargs.items()}
                # add any keyword args that aren't already in kwargs
                for col in arg.get_columns():
                    if col._short_name and col._short_name.lower() not in lower_case_kwargs:
                        kwargs[col._short_name] = col(arg._table)
            else:
                clean_args.append(arg)
        for k in list(kwargs.keys()):
            concept._dot(k)
        super().__init__(concept, clean_args, kwargs)
        self._identity_only = identity_only
        self._row_ids = row_ids

#------------------------------------------------------
# FilterBy
#------------------------------------------------------

class FilterBy(Expression):
    def __init__(self, item: Concept|Ref, kwargs: dict[str, Any]):
        concept = item._concept if isinstance(item, Ref) else item
        super().__init__(concept, [item], kwargs)

#------------------------------------------------------
# FieldRef
#------------------------------------------------------

class FieldRef(Variable):
    def __init__(self, root: Chain|Relationship|Expression, field: str|int|Concept, resolved: Field, resolved_ix:int):
        super().__init__(root._model)
        self._root = root
        self._field = field
        self._resolved = resolved
        self._resolved_ix = resolved_ix

    def _to_concept(self) -> Concept:
        return self._resolved.type

    def __getattr__(self, item):
        if item.startswith("_"):
            return object.__getattribute__(self, item)
        return Chain(self, self._resolved.type._dot(item))

    def to_dict_key(self) -> tuple:
        return (self._root.to_dict_key(), self._resolved._id)


#------------------------------------------------------
# MetaRef
#------------------------------------------------------

class MetaRef(Variable):
    def __init__(self, target: Concept|Relationship|Field):
        model = target._model if isinstance(target, DSLBase) else target.type._model
        super().__init__(model)
        self._target = target

#------------------------------------------------------
# TupleVariable
#------------------------------------------------------

class TupleVariable:
    def __init__(self, items: Sequence[Value|Distinct]):
        self._items = list(items)
        self._source = SourcePos.new()

#------------------------------------------------------
# AsBool
#------------------------------------------------------

class AsBool(Variable):
    def __init__(self, item: Variable):
        super().__init__(item._model)
        self._item = item

    def _to_concept(self) -> Concept:
        return CoreConcepts["Boolean"]

#------------------------------------------------------
# Alias
#------------------------------------------------------

class Alias(Variable):
    def __init__(self, source: Variable, alias: str):
        super().__init__(source._model)
        self._source = source
        self._alias = alias

    def _to_concept(self) -> Concept|None:
        return self._source._to_concept()

    def __format__(self, format_spec: str) -> str:
        if isinstance(self._source, Concept):
            if format_spec:
                exc("Invalid alias", f"Alias already specifies an alias for this concept, you can remove `:{format_spec}`")
            return self._source.__format__(self._alias)
        return super().__format__(format_spec)

#------------------------------------------------------
# Match
#------------------------------------------------------

class Match(DerivedTable):
    def __init__(self, model:Model, *items: Statement):
        super().__init__(model)
        t = type(self)
        self._items = [
            x
            for item in items
            for x in (item._items if (type(item) is t and isinstance(item, Match)) else (item,))
        ] # flatten nested Matches/Unions

    def _get_cols(self) -> list[DerivedColumn]:
        return [DerivedColumn(self, i) for i in range(self._arg_count())]

    def _arg_count(self) -> int:
        counts = []
        for item in self._items:
            if isinstance(item, DerivedTable):
                try:
                    counts.append(len(item._columns))
                except ValueError:
                    counts.append(0)
            else:
                # Expressions with no output and Not are filters, do not count as returning values
                is_filter = isinstance(item, Not) or (isinstance(item, Expression) and not item._has_output)
                counts.append(0 if is_filter else 1)
        if not counts:
            return 0
        first = counts[0]
        if any(c != first for c in counts[1:]):
            exc("Inconsistent branches",
                f"All branches in a {self.__class__.__name__} must have the same number of returned values",
                [source(self)])
        return first

    def __getattr__(self, item) -> Chain:
        if item.startswith("_"):
            return object.__getattribute__(self, item)
        return getattr(self._columns[-1], item)

#------------------------------------------------------
# Union
#------------------------------------------------------

class Union(Match):
    def __init__(self, model:Model, *items: Value):
        super().__init__(model, *items)

#------------------------------------------------------
# Not
#------------------------------------------------------

class Not(DSLBase):
    def __init__(self, *items: Value, model:Model):
        super().__init__(model)
        self._items = items

    def __or__(self, other) -> Match:
        return Match(self._model, self, other)

    def __and__(self, other) -> Fragment:
        if isinstance(other, Fragment):
            new = other.where()
            new._where.insert(0, self)
            return new
        return self._model.where(self, other)

#------------------------------------------------------
# Distinct
#------------------------------------------------------

class Distinct(DSLBase):
    def __init__(self, *items: Value, model:Model):
        super().__init__(model)
        self._items = items

#------------------------------------------------------
# Aggregate
#------------------------------------------------------

class Group(DSLBase):
    def __init__(self, *args: Value):
        model = [arg._model for arg in args if isinstance(arg, Variable)][0] if args else None
        super().__init__(model) # type: ignore
        self._args = list(args)

    def _extend(self, args: Sequence[Value]) -> Group:
        new = Group(*self._args)
        new._args.extend(args)
        return new

    def _clone(self):
        return Group(*self._args)

class Aggregate(Variable):
    def __init__(self, op: Relationship, *args: Value|Distinct, check_args: bool = True, distinct: bool = False):
        model = self._find_model(args) or op._model
        super().__init__(model)
        self._op = op
        self._where = Fragment(model)
        self._group = Group()
        self._args: list[Value] = []
        self._projection_args: list[Value] = []
        self._distinct = distinct
        if check_args:
            # unwrap distinct if present
            if any(isinstance(arg, Distinct) for arg in args):
                if len(args) != 1:
                    exc("Invalid distinct", "Distinct must be applied to all arguments", [source(self)])
                assert isinstance(args[0], Distinct)
                args = args[0]._items
                self._distinct = True

            args = cast(tuple[Value], args)

            num_inputs = sum(f.is_input for f in op._fields)
            if len(args) < num_inputs:
                need = [f.name for f in op._fields if f.is_input][len(args):]
                exc("Missing argument",
                    f"`{op._short_name or 'Relationship'}(..)` is missing: {', '.join(need)}",
                    [source(self)])

            self._projection_args = list(args[:-num_inputs] if num_inputs else args)
            supplied = iter(args[-num_inputs:] if num_inputs else [])

            self._args = [
                (next(supplied) if f.is_input else f.type.ref(f.name))
                for f in op._fields
            ]

    def _find_model(self, args: Sequence[Value|Distinct|TupleVariable]) -> Model|None:
        for arg in args:
            if isinstance(arg, (Variable, Distinct)):
                return arg._model
            elif isinstance(arg, TupleVariable):
                for item in arg._items:
                    if isinstance(item, (Variable, Distinct)):
                        return item._model
        return None


    def where(self, *args: Value) -> Aggregate:
        new = self._clone()
        new._where = new._where.where(*args)
        return new

    def per(self, *args: Value) -> Aggregate:
        new = self._clone()
        new._group = new._group._extend(args)
        return new

    def _clone(self):
        agg = Aggregate(self._op, check_args=False)
        agg._args = self._args
        agg._projection_args = self._projection_args
        agg._where = self._where
        agg._group = self._group
        agg._distinct = self._distinct
        return agg

#------------------------------------------------------
# Data
#------------------------------------------------------

class Data(Table):
    def __init__(self, df:DataFrame, model:Model):
        schema = {}
        for col in df.columns:
            _type = df[col].dtype
            if pd.api.types.is_datetime64_any_dtype(_type):
                col_type = "DateTime"
            elif pd.api.types.is_object_dtype(_type) and self._is_date_column(df[col]):
                col_type = "Date"
            else:
                col_type = pytype_to_concept_name.get(_type, "Any")
            if isinstance(col, int):
                col = f"col{col}"
            schema[col] = CoreConcepts[col_type]
        super().__init__("Data", schema, model)
        self._data = df

    def _is_date_column(self, col) -> bool:
        sample = col.dropna()
        if sample.empty:
            return False
        sample_value = sample.iloc[0]
        return isinstance(sample_value, dt.date) and not isinstance(sample_value, dt.datetime)

    @staticmethod
    def raw_to_df(data: DataFrame | list[tuple] | list[dict], columns:list[str]|None) -> DataFrame:
        if isinstance(data, DataFrame):
            return data
        if not data:
            return DataFrame()
        if isinstance(data, list):
            if isinstance(data[0], tuple):
                # Named tuple check
                if hasattr(data[0], '_fields'):
                    return DataFrame([t._asdict() for t in data]) #type: ignore
                return DataFrame(data, columns=columns)
            elif isinstance(data[0], dict):
                return DataFrame(data)
        raise TypeError(f"Cannot convert {type(data)} to DataFrame. Use DataFrame, list of tuples, or list of dicts.")

#------------------------------------------------------
# Enum
#------------------------------------------------------

class ModelEnumMeta(EnumMeta):
    _concept: Concept
    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_") or isinstance(value, self):
            super().__setattr__(name, value)
        elif isinstance(value, (Relationship, Reading)):
            setattr(self._concept, name, value)
        else:
            raise AttributeError(f"Cannot set attribute {name} on {type(self).__name__}")

    def __format__(self, format_spec: str) -> str:
        return format(self._concept, format_spec)

class ModelEnum(Enum, metaclass=ModelEnumMeta):
    _model:Model

    def __init__(self, *args) -> None:
        super().__init__(*args)
        self._source = SourcePos.new()

    def _compile_lookup(self):
        concept = getattr(self.__class__, "_concept")
        return concept.to_identity(name=self.name)

    @classmethod
    def lookup(cls, value:Variable|str):
        concept = cls._concept
        return concept.to_identity(name=value)

    # Python 3.10 doesn't correctly populate __members__ by the time it calls
    # __init_subclass__, so we need to initialize the members lazily when we
    # encounter the enum for the first time.
    @classmethod
    def _init_members(cls):
        if cls._has_inited_members:
            return
        c = cls._concept
        # Add the name and value attributes to the hashes we create for the enum
        members = [
            c.new(name=name, value=value.value)
            for name, value in cls.__members__.items()
        ]
        cls._model.define(*members)
        cls._has_inited_members = True

    def __format__(self, format_spec: str) -> str:
        return format(self._concept, format_spec)

def create_enum_class(model: Model):
    class AttachedModelEnum(ModelEnum):
        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__(**kwargs)
            # this is voodoo black magic that is doing meta meta programming where
            # we are plugging into anytime a new subtype of this class is created
            # and then creating a concept to represent the enum. This happens both
            # when you do `class Foo(Enum)` and when you do `Enum("Foo", [a, b, c])`
            cls._model = model
            c = model.Concept(
                cls.__name__,
                extends=[CoreConcepts["Enum"]],
                identify_by={"name": CoreConcepts["String"]}
            )
            model.enums.append(cls)
            model.enums_index[cls.__name__] = cls
            cls._has_inited_members = False
            cls._concept = c

    return AttachedModelEnum

#------------------------------------------------------
# Value
#------------------------------------------------------

Value = Variable|Primitive|Field|TupleVariable

#------------------------------------------------------
# Statement
#------------------------------------------------------

Statement = Value | Group | Not | Distinct | Aggregate
"""Union of statement types."""

StatementAndSchema = Statement | TableSchema

#------------------------------------------------------
# Fragment
#------------------------------------------------------

@include_in_docs
class Fragment(DerivedTable):
    """Composable chunk of a query with select/where/define state.

    Parameters
    ----------
    model : Model
        The semantic model that provides type information and stores any
        definitions produced by this fragment.
    parent : Fragment, optional
        An existing fragment to inherit selection and filter state from when
        building chained queries.
    """

    def __init__(self, model:Model, parent:Fragment|None=None):
        super().__init__(model)
        self._id = next(_global_id)
        self._select = []
        self._where = []
        self._require = []
        self._define = []
        self._order_by = []
        self._limit = 0
        self._model = model
        self._into:Optional[Table] = None
        self._is_into_update = False
        self._has_executed = False
        assert self._model, "Fragment must have a model"

        self._parent = parent
        # self._source = runtime_env.get_source_pos()
        self._meta = {}
        self._annotations = []
        if parent is not None:
            self._select.extend(parent._select)
            self._where.extend(parent._where)
            self._require.extend(parent._require)
            self._define.extend(parent._define)
            self._order_by.extend(parent._order_by)
            self._limit = parent._limit


    def _add_items(self, items:Sequence[Statement], to_attr:list[Statement]):
        # TODO: ensure that you are _either_ a select, require, or then
        # not a mix of them
        model = self._model

        # remove any existing rules that this consumes
        for item in itertools.chain(items, [self._parent]):
            if isinstance(item, Fragment) and item._is_effect():
                model._remove_rule(item)

        to_attr.extend(items)
        if self._is_effect():
            model._add_rule(self)

        return self

    #------------------------------------------------------
    # Select arg handling
    #------------------------------------------------------

    def _check_select_args(self, args:Sequence[StatementAndSchema]) -> Sequence[Statement]:
        clean_args = []
        for arg in args:
            # If you select x > y, treat that as AsBool(x > y)
            if isinstance(arg, Expression) and not arg._has_output:
                clean_args.append(AsBool(arg))
            elif isinstance(arg, TableSchema):
                for col in arg.get_columns():
                    clean_args.append(col(arg._table))
            else:
                clean_args.append(arg)
        return clean_args

    #------------------------------------------------------
    # Core API
    #------------------------------------------------------

    def where(self, *args: Statement) -> Fragment:
        f = Fragment(self._model, parent=self)
        return f._add_items(args, f._where)

    def select(self, *args: StatementAndSchema) -> Fragment:
        f = Fragment(self._model, parent=self)
        return f._add_items(self._check_select_args(args), f._select)

    def require(self, *args: Statement) -> Fragment:
        f = Fragment(self._model, parent=self)
        return f._add_items(args, f._require)

    def define(self, *args: Statement) -> Fragment:
        f = Fragment(self._model, parent=self)
        return f._add_items(args, f._define)

    if not TYPE_CHECKING:
        def order_by(self, *args: Any) -> Fragment:
            exc("Feature unavailable", "The 'order_by' method is not yet available when querying RAI directly.", [
                source(self),
                "You can use [cyan]`relationalai.semantics.std.aggregates.rank`[/cyan] and select it as the first column as a temporary substitute.",
            ])
            f = Fragment(self._model, parent=self)
            return f._add_items(args, f._order_by)

        def limit(self, n:int) -> Fragment:
            exc("Feature unavailable", "The 'order_by' method is not yet available when querying RAI directly.", [
                source(self),
                "You can use [cyan]`relationalai.semantics.std.aggregates.limit`[/cyan] in a where clause as a temporary substitute.",
            ])
            f = Fragment(self._model, parent=self)
            f._limit = n
            return f

    # def meta(self, **kwargs: Any) -> Fragment:
    #     self._meta.update(kwargs)
    #     return self

    def annotate(self, *annos:Expression|Relationship) -> Fragment:
        self._annotations.extend(annos)
        return self

    #------------------------------------------------------
    # into
    #------------------------------------------------------

    def into(self, table: Table, update=False) -> Fragment:
        f = Fragment(self._model, parent=self)
        f._into = table
        f._is_into_update = update
        self._model.exports.add(f)
        return f

    #------------------------------------------------------
    # Execution
    #------------------------------------------------------

    def exec(self):
        if self._has_executed:
            return
        if self._into is None:
            exc("Cannot execute", "Query must have an 'into' table specified to execute.", [source(self)])
        from relationalai.shims.executor import execute
        self._has_executed = True
        return execute(self, self._model, export_to=self._into._name, update=self._is_into_update)

    def to_df(self):
        from relationalai.shims.executor import execute
        return execute(self, self._model)

    def inspect(self):
        print(self.to_df())

    #------------------------------------------------------
    # helpers
    #------------------------------------------------------

    def _is_effect(self) -> bool:
        return bool(self._define or self._require)

    def _is_where_only(self) -> bool:
        return not self._select and not self._define and not self._require and not self._order_by

    #------------------------------------------------------
    # And/Or
    #------------------------------------------------------

    def __or__(self, other) -> Match:
        return Match(self._model, self, other)

    def __and__(self, other) -> Fragment:
        if not isinstance(other, Fragment):
            other = Fragment(self._model).where(other)
        if self._is_where_only() and other._is_where_only():
            return self.where(*other._where)
        elif self._is_where_only():
            return other.where(*self._where)
        elif other._is_where_only():
            return self.where(*other._where)
        else:
            raise Exception("Cannot AND two non-where-only fragments")

    #------------------------------------------------------
    # DerivedTable cols
    #------------------------------------------------------

    def _get_cols(self):
        return [DerivedColumn.from_value(self, i, col) for i, col in enumerate(self._select)]

    #------------------------------------------------------
    # Marterialize
    #------------------------------------------------------

    def to_metamodel(self):
        m = self._model
        return m._compiler.compile(self)

#------------------------------------------------------
# Model
#------------------------------------------------------

@include_in_docs
class Model:
    """Class representing a semantic model.

    Parameters
    ----------
    name : str, optional
        Name of the model.
    exclude_core : bool, optional
        If True, the core library will not be included by default.
    is_library : bool, optional
        If True, this model is a library and will not be added to the global
        list of models.

    Examples
    --------
    Create a model object:

    >>> from relationalai.semantics import Model
    >>> model = Model(name="MyModel")
    """
    all_models:list[Model] = []

    def __init__(self, name: str = "", exclude_core: bool = False, is_library: bool = False):
        self.name = name
        self.defines:KeyedSet[Fragment] = KeyedSet(dsl_key)
        self.requires:KeyedSet[Fragment] = KeyedSet(dsl_key)
        self.exports:KeyedSet[Fragment] = KeyedSet(dsl_key)
        self.libraries = []
        self.concepts: list[Concept] = []
        self.tables: list[Table] = []
        self.relationships: list[Relationship] = []
        self.enums: list[Type[ModelEnum]] = []
        self.concept_index: dict[str, Concept] = {}
        self.table_index: dict[str, Table] = {}
        self.relationship_index: dict[str, Relationship] = {}
        self.enums_index: dict[str, Type[ModelEnum]] = {}
        self._source = SourcePos.new()

        self.Enum = create_enum_class(self)

        if not is_library:
            Model.all_models.append(self)

        self.__compiler = None
        if not exclude_core:
            self.libraries.append(CoreLibrary)

    #------------------------------------------------------
    # Internal
    #------------------------------------------------------

    @property
    def _compiler(self):
        if not self.__compiler:
            from .front_compiler import FrontCompiler
            self.__compiler = FrontCompiler(self)
        return self.__compiler

    def _find_concept(self, name: str) -> Concept|None:
        if name in self.concept_index:
            return self.concept_index[name]
        if name in self.table_index:
            return self.table_index[name]
        for lib in self.libraries:
            found = lib._find_concept(name)
            if found is not None:
                return found

    def _remove_rule(self, fragment: Fragment) -> None:
        if fragment._define and fragment in self.defines:
            self.defines.remove(fragment)
        elif fragment._require and fragment in self.requires:
            self.requires.remove(fragment)
        else:
            raise ValueError("Fragment must have either define or require clauses to be removed as a rule")

    def _add_rule(self, fragment: Fragment) -> None:
        if fragment._define and fragment not in self.defines:
            self.defines.add(fragment)
        elif fragment._require and fragment not in self.requires:
            self.requires.add(fragment)
        else:
            raise ValueError("Fragment must have either define or require clauses to be added as a rule")

    #------------------------------------------------------
    # Primary API
    #------------------------------------------------------

    def Concept(self, name: str, extends: list[Concept] = [], identify_by: dict[str, Property|Concept] = {}) -> Concept:
        c = Concept(name, extends=extends, identify_by=identify_by, model=self)
        self.concepts.append(c)
        self.concept_index[name] = c
        return c

    def Table(self, path: str, schema: dict[str, Concept] = {}) -> Table:
        t = Table(path, schema=schema, model=self)
        self.tables.append(t)
        self.table_index[path] = t
        return t

    def Relationship(self, reading: str = "", fields: list[Field] = [], short_name: str = "") -> Relationship:
        r = Relationship(self, reading, fields, short_name)
        self.relationships.append(r)
        self.relationship_index[short_name] = r
        return r

    def Property(self, reading: str = "", fields: list[Field] = [], short_name: str = "") -> Property:
        p = Property(self, reading, fields, short_name)
        self.relationships.append(p)
        self.relationship_index[short_name] = p
        return p

    def select(self, *args: StatementAndSchema) -> Fragment:
        f = Fragment(self)
        return f.select(*args)

    def where(self, *args: Statement) -> Fragment:
        f = Fragment(self)
        return f.where(*args)

    def require(self, *args: Statement) -> Fragment:
        f = Fragment(self)
        return f.require(*args)

    def define(self, *args: Statement) -> Fragment:
        f = Fragment(self)
        return f.define(*args)

    def union(self, *items: Value) -> Union:
        return Union(self, *items)

    def data(self, data: DataFrame | list[tuple] | list[dict], columns: list[str]|None = None) -> Data:
        df = Data.raw_to_df(data, columns)
        return Data(df, self)

    def not_(self, *items: Value) -> Not:
        return Not(*items, model=self)

    def distinct(self, *items: Value) -> Distinct:
        return Distinct(*items, model=self)

    #------------------------------------------------------
    # Meta
    #------------------------------------------------------

    def to_metamodel(self) -> mModel:
        return self._compiler.compile_model(self)

#------------------------------------------------------
# Library
#------------------------------------------------------

class Library(Model):
    def __init__(self, name: str, exclude_core: bool = False):
        super().__init__(name=name, exclude_core=exclude_core, is_library=True)

    def Type(self, name: str, super_types: list[Concept] = []) -> Concept:
        c = self.Concept(name, extends=super_types)
        self._register_builtin(c)
        return c

    def Relation(self, name: str, fields: list[Field], overloads: list[list[Concept]]|None = None, annotations: list[Any]|None = None) -> Relationship:
        r = Relationship(self, "", fields, name, allow_no_fields=True, overloads=overloads)
        self.relationships.append(r)
        self.relationship_index[name] = r
        self._register_builtin(r)
        # TODO - deal with annotations
        return r

    def _register_builtin(self, builtin: Concept|Relationship) -> None:
        """ Register a builtin concept or relationship in the global builtins registry. """
        builtins.register(self.name, self._compiler.to_relation(builtin))


#------------------------------------------------------
# Core Library
#------------------------------------------------------

# Library is populated when relationalai.semantics.frontend.core is imported
CoreLibrary: Library = Library("core", exclude_core=True)
CoreRelationships = CoreLibrary.relationship_index
CoreConcepts = CoreLibrary.concept_index


#------------------------------------------------------
# todo
#------------------------------------------------------

"""
x new/to_identity/filter_by
x capture all defines/requires
x identify_by
x figure out Table/DerivedTable
x data node
x in_
x error on invalid iteration/sum/min/max
x Aggregates
x fill in std.agg functions when Thiago's lib stuff lands
- Error?
x rank stuff
- validation checks for identify_by
- unique
- Enum?
"""



# For builder components.
from v0.relationalai.semantics import Integer, define, sum, not_
from v0.relationalai.experimental.paths.graph import Graph
from v0.relationalai.experimental.paths.path_algorithms.one_sided_ball_upto import ball_upto
from v0.relationalai.experimental.paths.path_algorithms.one_sided_ball_repetition import ball_with_repetition


def compute_usp(g: Graph, Source, Target, max_length=None):
    # Computes the Union of Shortest Paths (USP) from a source set to a destination set
    # USP is a subgraph of the original graph

    edge = g.Edge
    Node = g.Node

    n, d = Integer.ref(), Integer.ref()
    tgt, u, v = Node.ref(), g.Node.ref(), g.Node.ref()

    ball = ball_upto(g, Source, Target, max_length)
    usp_nodes = g.model.Relationship(f"usp_nodes {{{Node}}}")
    usp = g.model.Relationship(f"usp {{{Node}}} {{{Node}}}")
    boundary = g.model.Relationship(f"usp_boundary {{{Node}}}")

    # propagate backwards to find the nodes that are in the USP:
    define(boundary(tgt)).where(
        Target(tgt),
        ball(n, tgt)
    )

    define(usp_nodes(tgt)).where(
        boundary(tgt)
    )

    define(usp_nodes(u)).where(
        usp(u, v)
    )

    define(usp(u, v)).where(
        usp_nodes(v),
        ball(d, v),
        ball(d - 1, u),
        edge(u, v)
    )

    return usp, boundary


def compute_uw(g: Graph, Source, Target, max_length):
    # Computes the Union of Walks (UW) up to a given length from a source set to a destination set

    edge = g.Edge
    Node = g.Node

    n = Integer.ref()
    tgt, u, v, w = Node.ref(), g.Node.ref(), g.Node.ref(), g.Node.ref()

    ball = ball_with_repetition(g, Source, max_length)
    uw = g.model.Relationship(f"uw {{{Node}}} {{Integer}} {{{Node}}}")
    boundary = g.model.Relationship(f"boundary {{{Node}}} {{Integer}}")

    define(uw(u, n, tgt)).where(
        Target(tgt),
        ball(n + 1, tgt),
        ball(n, u),
        edge(u, tgt)
    )

    define(uw(u, n, v)).where(
        uw(v, n + 1, w),
        ball(n, u),
        edge(u, v)
    )

    define(boundary(tgt, n)).where(
        Target(tgt),
        ball(n, tgt),
        not_(uw(tgt, n, u))
    )

    return uw, boundary


def compute_nsp_from_usp(g: Graph, usp, Source, Target, Boundary):
    Node = g.Node

    n = Integer.ref()
    tgt, u, v = Node.ref(), Node.ref(), Node.ref()

    nsp = g.model.Relationship(f"num_shortest {{{Node}}} {{Integer}}")

    define(nsp(tgt, 1)).where(
        Boundary(tgt)
    )

    define(nsp(tgt, 1)).where(
        Target(tgt),
        Source(tgt)
    )

    define(nsp(u, sum(v, n).per(u))).where(
        nsp(v, n),
        usp(u, v)
    )

    return nsp


def compute_nw_from_uw(g: Graph, usp, Target, Boundary):
    Node = g.Node

    n, m = Integer.ref(), Integer.ref()
    tgt, u, v = Node.ref(), Node.ref(), Node.ref()

    nw = g.model.Relationship(f"num_walks {{{Node}}} {{Integer}} {{Integer}}")

    define(nw(tgt, n, 1)).where(
        Boundary(tgt, n)
    )

    define(nw(u, n, sum(v, n + 1, m).per(u, n))).where(
        nw(v, n + 1, m),
        usp(u, n, v),
        not_(Target(u))
    )

    define(nw(u, n, 1 + sum(v, n + 1, m).per(u, n))).where(
        nw(v, n + 1, m),
        usp(u, n, v),
        Target(u),
        not_(Boundary(u, n))
    )

    return nw


def compute_nsp(g: Graph, Source, Target):
    # Computes the number of shortest paths (NSP) from a source set to a destination set

    usp, Boundary  = compute_usp(g, Source, Target)
    nsp = compute_nsp_from_usp(g, usp, Source, Target, Boundary)

    return nsp


def compute_nw(g: Graph, Source, Target, max_length):
    # Computes the number of shortest paths (NSP) from a source set to a destination set

    uw, Boundary  = compute_uw(g, Source, Target, max_length)
    nw = compute_nw_from_uw(g, uw, Target, Boundary)

    return nw



from __future__ import annotations
import os

from .generic import GenericEnvironment

class CIEnvironment(GenericEnvironment):
    @classmethod
    def detect(cls):
        """Detect if we're running in a CI environment."""
        # Check for actual CI environments (not just environment variables that might be set locally)
        if os.getenv("CI") == "true" and (
            os.getenv("GITHUB_ACTIONS") == "true" or
            os.getenv("GITLAB_CI") == "true" or
            os.getenv("JENKINS_URL") or
            os.getenv("TRAVIS") == "true" or
            os.getenv("CIRCLECI") == "true"
        ):
            return True

        # Check for GitHub Actions specifically (more reliable than GITHUB_REF_NAME)
        if os.getenv("GITHUB_ACTIONS") == "true":
            return True

        return False



import warnings
from v0.relationalai.semantics.sql import sql, Compiler

__all__ = ["sql", "Compiler"]

warnings.warn(
    "relationalai.early_access.sql.* is deprecated. "
    "Please migrate to relationalai.semantics.sql.*",
    DeprecationWarning,
    stacklevel=2,
)


import warnings

from v0.relationalai.semantics.metamodel.typer.typer import to_type, is_primitive, to_base_primitive, _NON_PARAMETRIC_PRIMITIVES

__all__ = ["to_type", "is_primitive", "to_base_primitive", "_NON_PARAMETRIC_PRIMITIVES"]

warnings.warn(
    "relationalai.early_access.metamodel.typer.typer is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.typer.typer",
    DeprecationWarning,
    stacklevel=2,
)


from collections import OrderedDict
from itertools import product
from types import MappingProxyType
from typing import Union, Sequence

import v0.relationalai.semantics as qb
from v0.relationalai.semantics import where
from v0.relationalai.early_access.dsl.bindings.common import Binding, RoleBinding, IdentifierConceptBinding, BindableTable, \
    BindableAttribute, SubtypeConceptBinding, ReferentConceptBinding
from v0.relationalai.early_access.dsl.codegen.binder import Binder
from v0.relationalai.early_access.dsl.codegen.common import BoundIdentifierConstraint, BoundRelationship
from v0.relationalai.early_access.dsl.codegen.relations import ValueMap, CompositeEntityMap, EntitySubtypeMap, \
    UnionEntityMap, AbstractEntityMap, SimpleConstructorEntityMap, ReferentEntityMap, InlineValueMap, \
    Map, MaterializedEntityMap, AbstractInlineEntityMap
from v0.relationalai.semantics.metamodel.util import OrderedSet

DEFAULT_WEAVER_CONFIG = MappingProxyType({
    'inline_value_maps': False,
    'inline_entity_maps': False,
})


class Weaver:

    def __init__(self, model, config=None):
        self._model = model
        self._config = config or DEFAULT_WEAVER_CONFIG
        self._reasoner = self._model.reasoner()
        self._binder = Binder(self._model)

        # Lookup dictionaries for generated relations
        self._binding_to_value_map: dict[Binding, Union[ValueMap, InlineValueMap]] = OrderedDict()
        self._ctor_binding_to_entity_map:\
            dict[IdentifierConceptBinding, Union[AbstractInlineEntityMap, MaterializedEntityMap]] = OrderedDict()
        self._ref_binding_to_entity_map:\
            dict[ReferentConceptBinding, Union[AbstractInlineEntityMap, MaterializedEntityMap]] = OrderedDict()
        self._bound_identifier_to_composite_entity_map:\
            dict[BoundIdentifierConstraint, OrderedSet[Union[AbstractInlineEntityMap, MaterializedEntityMap]]] = OrderedDict()
        self._subtype_binding_to_entity_map:\
            dict[Union[IdentifierConceptBinding, ReferentConceptBinding], Union[AbstractInlineEntityMap, MaterializedEntityMap]] = OrderedDict()
        self._filtering_subtype_binding_to_entity_map:\
            dict[SubtypeConceptBinding, Union[AbstractInlineEntityMap, MaterializedEntityMap]] = OrderedDict()
        self._supertype_to_entity_map: dict[qb.Concept, UnionEntityMap] = OrderedDict()

    def generate(self):
        self._generate_value_maps()
        self._generate_entity_maps()
        self._generate_semantic_predicates()

    def _generate_value_maps(self):
        for binding in self._binder.value_type_bindings():
            self._generate_value_map(binding)

    def _generate_value_map(self, binding: Binding):
        role = self._binder.lookup_binding_role(binding)
        if self._cfg_should_inline_value_maps():
            value_map = InlineValueMap(self._model, binding, role)
        else:
            value_map = ValueMap(self._model, binding, role)
        self._binding_to_value_map[binding] = value_map

    def _look_up_value_map(self, binding: Binding):
        """
        Look up a value map in the model by value type binding.
        """
        return self._binding_to_value_map[binding]

    def _generate_entity_maps(self):
        self._generate_simple_ctor_entity_maps()
        self._generate_referent_entity_maps()
        self._generate_composite_entity_maps()
        self._generate_subtype_entity_maps()
        self._generate_filtered_subtype_entity_maps()
        self._generate_supertype_entity_maps()

    def _generate_simple_ctor_entity_maps(self):
        for binding in self._binder.constructor_bindings():
            self._generate_simple_ctor_entity_map(binding)

    def _generate_simple_ctor_entity_map(self, binding: IdentifierConceptBinding):
        ctor_role = self._binder.lookup_binding_role(binding)
        role = ctor_role.sibling()
        assert role is not None, "Constructor binding must have a sibling role"
        identifier_to_role_map = self._lookup_supertype_identifier_role_maps(binding)

        entity_map = SimpleConstructorEntityMap(self._model, binding, role, identifier_to_role_map)
        entity_map = self._try_materialize_entity_map(entity_map)
        self._ctor_binding_to_entity_map[binding] = entity_map

    def _lookup_supertype_identifier_role_maps(self, binding: IdentifierConceptBinding):
        """
        Build a mapping from each reference scheme of the binding's concept to the corresponding constructing role map.
        """
        concept = binding.entity_type
        ref_scheme_to_role_map = OrderedDict()

        for ref_scheme in self._reasoner.ref_schemes_of(concept):
            constructed_concept = self._reasoner.identifies_concept(ref_scheme)

            #=
            # May need to look up the constructor binding for another concept with (.id) in the concept's hierarchy.
            # E.g., if the concept is a subtype with multiple ref schemes, we need to look up the respective constructor
            # binding for each supertype with a reference scheme.
            #=
            constructor_binding = (
                binding if constructed_concept is concept
                else self._binder.lookup_constructor_binding_by_source(constructed_concept, binding.column.table)
            )

            ctor_role = self._binder.lookup_binding_role(constructor_binding)
            player = ctor_role.player()

            # If the constructor role is played by a primitive, use its sibling's player
            if player._is_primitive():
                sibling = ctor_role.sibling()
                assert sibling is not None, "Constructor roles must have a sibling role"
                player = sibling.player()

            role_map = self._lookup_constructing_role_map(constructor_binding, player)
            ref_scheme_to_role_map[ref_scheme] = role_map
        return ref_scheme_to_role_map

    def _lookup_constructing_role_map(self, binding: Union[IdentifierConceptBinding, ReferentConceptBinding],
                                      concept: qb.Concept):
        if self._binder.is_value_type_binding(binding):
            role_map = self._binding_to_value_map[binding]
        elif isinstance(binding, IdentifierConceptBinding):
            #=
            # This call is needed to ensure we look up the ctor binding for the concept, which may not be the same as
            # the one in the binding. This is because the reference scheme may have an entity type in the constructing
            # role, hence we need to look up the constructor binding and then the role map for that binding.
            #=
            ctor_binding = self._binder.lookup_constructor_binding(concept, binding.column)
            assert isinstance(ctor_binding, IdentifierConceptBinding)
            role_map = self._ctor_binding_to_entity_map[ctor_binding]
        elif isinstance(binding, ReferentConceptBinding):
            # TODO: check why we need `binder.lookup_constructor_binding` above?
            role_map = self._ref_binding_to_entity_map[binding]
        else:
            raise ValueError(f'Cannot lookup constructing role map for binding {binding} and concept {concept}')
        return role_map

    def _generate_referent_entity_maps(self):
        for binding in self._binder.referent_concept_bindings():
            self._generate_referent_entity_map(binding)

    def _generate_referent_entity_map(self, binding: ReferentConceptBinding):
        ctor_role = self._binder.lookup_binding_role(binding)
        role = ctor_role.sibling()
        assert role is not None, "Referent concept binding must have a sibling role"
        constructing_role_map = self._lookup_constructing_role_map(binding, binding.entity_type)
        entity_map = ReferentEntityMap(self._model, binding, role, constructing_role_map)
        entity_map = self._try_materialize_entity_map(entity_map)
        self._ref_binding_to_entity_map[binding] = entity_map

    def _generate_composite_entity_maps(self):
        for constraint in self._binder.bound_identifier_constraints():
            if constraint.is_external():
                # generate a composite entity map
                self._generate_composite_entity_map(constraint)

    def _generate_composite_entity_map(self, constraint: BoundIdentifierConstraint):
        role_map_combinations = self._composite_entity_map_combinations(constraint)
        entity_concept = constraint.concept
        for role_maps in role_map_combinations:
            entity_map = CompositeEntityMap(self._model, entity_concept, *role_maps)
            entity_map = self._try_materialize_entity_map(entity_map)
            self._bound_identifier_to_composite_entity_map.setdefault(constraint, OrderedSet()).add(entity_map)

    def _composite_entity_map_combinations(self, bound_uc: BoundIdentifierConstraint) -> Sequence[Sequence[Map]]:
        role_bindings = bound_uc.role_bindings
        # Build a list of binding lists for each role in the right order
        role_binding_lists = [role_bindings[role] for role in bound_uc.constraint.roles()]
        # Generate all possible combinations of bindings and convert each combination to a list
        all_combinations = [list(combination) for combination in product(*role_binding_lists)]  # pyright: ignore[reportArgumentType, reportCallIssue]
        # Construct entity map combinations using the _ref_entity_map method
        role_map_combinations = [
            [self._lookup_ref_role_map(binding) for binding in combination]
            for combination in all_combinations
        ]
        return role_map_combinations

    def _lookup_ref_role_map(self, binding: RoleBinding):
        concept = binding.role.player()
        if concept._is_primitive():
            role_map = self._binding_to_value_map[binding]
        else:
            ctor_binding = self._binder.lookup_constructor_binding(concept, binding.column)
            role_map = self._lookup_ref_role_map_by_ctor_binding(ctor_binding)
        return role_map

    def _lookup_ref_role_map_by_ctor_binding(self, binding: Union[IdentifierConceptBinding, ReferentConceptBinding]):
        if isinstance(binding, IdentifierConceptBinding):
            role_map = self._ctor_binding_to_entity_map[binding]
        elif isinstance(binding, ReferentConceptBinding):
            role_map = self._ref_binding_to_entity_map[binding]
        else:
            raise ValueError(f'Cannot lookup role map for binding {binding} - invalid type of binding')
        return role_map

    def _lookup_ref_entity_maps_by_column(self, concept: qb.Concept, column: BindableAttribute):
        if self._reasoner.is_composite_concept(concept):  # composite entity type
            identifier_constraint = self._binder.lookup_bound_identifier_constraint_by_column(concept, column)
            entity_map_candidates = self._bound_identifier_to_composite_entity_map[identifier_constraint]
            # note: for now, we look up by table, though needs fixing
            return self._find_matching_entity_maps(entity_map_candidates, concept, column.table)
        elif self._reasoner.has_own_ref_scheme(concept):  # simple entity type
            entity_map_candidates = (list(self._ctor_binding_to_entity_map.values()) +
                                     list(self._ref_binding_to_entity_map.values()))
            return self._find_matching_entity_maps(entity_map_candidates, concept, column)
        elif self._reasoner.is_exclusive_entity_type(concept):  # exclusive entity type
            entity_map_candidates = self._supertype_to_entity_map.values()
            # note: for now, we look up by table, though needs fixing
            return self._find_matching_entity_maps(entity_map_candidates, concept, column.table, strict_match=True)
        else:  # entity subtype
            entity_map_candidates = self._subtype_binding_to_entity_map.values()
            return self._find_matching_entity_maps(entity_map_candidates, concept, column)

    def _lookup_ref_entity_maps_by_source(self, concept: qb.Concept, source: BindableTable):
        if self._reasoner.is_composite_concept(concept):  # composite entity type
            identifier_constraints = self._binder.lookup_bound_identifier_constraint_by_source(concept, source)
            entity_map_candidates = [entity_map for identifier_constraint in identifier_constraints
                                     for entity_map in
                                     self._bound_identifier_to_composite_entity_map[identifier_constraint]]
            return self._find_matching_entity_maps(entity_map_candidates, concept, source)
        elif self._reasoner.has_own_ref_scheme(concept):  # simple entity type
            entity_map_candidates = (list(self._ctor_binding_to_entity_map.values()) +
                                     list(self._ref_binding_to_entity_map.values()))
            return self._find_matching_entity_maps(entity_map_candidates, concept, source)
        elif self._reasoner.is_exclusive_entity_type(concept):  # exclusive entity type
            entity_map_candidates = self._supertype_to_entity_map.values()
            return self._find_matching_entity_maps(entity_map_candidates, concept, source, strict_match=True)
        else:  # entity subtype
            entity_map_candidates = (list(self._subtype_binding_to_entity_map.values()) +
                                     list(self._filtering_subtype_binding_to_entity_map.values()))
            return self._find_matching_entity_maps(entity_map_candidates, concept, source)
        # TODO : revise the above logic and see if uniqueness check is needed for anythine else

    @staticmethod
    def _find_matching_entity_maps(entity_maps, concept: qb.Concept,
                                   source_or_col: Union[BindableTable, BindableAttribute],
                                   strict_match: bool=False):
        strictly_matching = OrderedSet()
        loosely_matching = OrderedSet()
        for entity_map in entity_maps:
            value_player = entity_map.value_player()
            if Weaver._matches_source(entity_map, source_or_col):
                if concept is value_player:
                    strictly_matching.add(entity_map)
                elif concept._isa(value_player):
                    loosely_matching.add(entity_map)
        matching_entity_maps = strictly_matching if bool(strictly_matching) or strict_match else loosely_matching
        if not matching_entity_maps:
            raise ValueError(f'No entity map found for concept {concept} and {source_or_col}')
        return matching_entity_maps

    @staticmethod
    def _matches_source(entity_map, source_or_col: Union[BindableTable, BindableAttribute]):
        if isinstance(source_or_col, BindableTable):
            return entity_map.table() == source_or_col
        elif isinstance(source_or_col, BindableAttribute):
            return entity_map.column() is source_or_col
        else:
            return False

    def _generate_subtype_entity_maps(self):
        for binding in self._binder.subtype_ctor_bindings():
            self._generate_subtype_entity_map(binding)

    def _generate_subtype_entity_map(self, binding: Union[IdentifierConceptBinding, ReferentConceptBinding]):
        if isinstance(binding, ReferentConceptBinding):
            ctor_entity_map = self._ref_binding_to_entity_map[binding]
        else:
            ctor_binding = self._binder.lookup_subtype_reference_binding(binding)
            ctor_entity_map = self._lookup_ref_role_map_by_ctor_binding(ctor_binding)
        entity_map = EntitySubtypeMap(self._model, binding, ctor_entity_map)
        entity_map = self._try_materialize_entity_map(entity_map)
        self._subtype_binding_to_entity_map[binding] = entity_map

    def _generate_filtered_subtype_entity_maps(self):
        for binding in self._binder.subtype_filtering_bindings():
            self._generate_filtered_subtype_entity_map(binding)

    def _generate_filtered_subtype_entity_map(self, binding: SubtypeConceptBinding):
        ctor_binding = self._binder.lookup_subtype_reference_binding(binding)
        ctor_entity_map = self._lookup_ref_role_map_by_ctor_binding(ctor_binding)
        entity_map = EntitySubtypeMap(self._model, binding, ctor_entity_map)
        entity_map = self._try_materialize_entity_map(entity_map)
        self._filtering_subtype_binding_to_entity_map[binding] = entity_map

    def _generate_supertype_entity_maps(self):
        simple_subtype_maps = self._subtype_binding_to_entity_map.values()
        filtering_subtype_maps = self._filtering_subtype_binding_to_entity_map.values()
        subtype_maps = (OrderedSet()
                        .update(simple_subtype_maps)
                        .update(filtering_subtype_maps))
        for subtype_map in subtype_maps:
            self._generate_supertype_entity_map(subtype_map)
        # after generating all supertype maps, try to materialize them
        [entity_map.materialize_population() for entity_map in self._supertype_to_entity_map.values()]

    def _generate_supertype_entity_map(self, subtype_map: AbstractEntityMap):
        subtype = subtype_map.value_player()
        supertype = self._reasoner.subtype_exclusive_supertype(subtype)
        if not supertype:
            return
        supertype_map = self._supertype_to_entity_map.get(supertype)
        if supertype_map is None:
            supertype_map = UnionEntityMap(self._model, supertype, subtype_map, generate_population=True)
            self._supertype_to_entity_map[supertype] = supertype_map
        else:
            supertype_map.update(subtype_map)
        # go higher in the hierarchy
        self._generate_supertype_entity_map(supertype_map)

    def _generate_semantic_predicates(self):
        for bound_relationship in self._binder.bound_relationships():
            self._generate_semantic_predicate(bound_relationship)

    def _generate_semantic_predicate(self, bound_relationship: BoundRelationship):
        if bound_relationship.relationship._unary():
            self._generate_unary_predicate_rule(bound_relationship)
        else:
            self._generate_nary_predicate_rule(bound_relationship)

    def _generate_unary_predicate_rule(self, bound_relationship: BoundRelationship):
        relationship = bound_relationship.relationship
        roles = relationship._roles()
        assert len(roles) == 1, "Unary predicate should have exactly one role"
        unary_role = roles[0]
        bound_concept = unary_role.player()
        if bound_concept._is_primitive():
            raise ValueError(f'Unary relationships cannot be defined on value types: {relationship}')
        role_bindings = bound_relationship.bindings
        for binding in role_bindings:
            if not self._binder.is_filtering_binding(binding):
                raise ValueError(f'Unary relationship roles must be bound by a filtering binding (use `filter_by`): {relationship}')
            self._generate_unary_predicate_rule_body(bound_concept, binding, relationship)

    def _generate_unary_predicate_rule_body(self, concept_player: qb.Concept, binding: Binding, relationship):
        entity_maps = self._lookup_ref_entity_maps_by_column(concept_player, binding.column)
        if not entity_maps:
            raise ValueError(
                f'Cannot generate semantic predicate for {relationship}, no role maps found')
        for entity_map in entity_maps:
            entity_ref, subformula_atoms = entity_map.formula()
            where(
                *subformula_atoms,
                where(binding.filter_by)
            ).define(relationship(entity_ref))

    def _generate_nary_predicate_rule(self, bound_relationship: BoundRelationship):
        relationship = bound_relationship.relationship
        roles = relationship._roles()
        #=
        # Each role should either be covered by a value map (value type role), or by an entity map (entity type role).
        # The entity map either can be inferred if unique exists for the entity type, or must be generated by
        # the respective EntityBinding.
        #=
        role_to_role_maps = OrderedDict()
        role_map_casts_to = OrderedDict()
        for role in roles:
            role_concept = role.player()
            if role_concept._is_primitive():
                # TODO: [SB] I suppose we can do bound_relationship.role_bindings[role] here?
                role_bindings = self._binder.lookup_role_bindings(role, bound_relationship.table)
                role_maps = [self._binding_to_value_map[binding] for binding in role_bindings]
            else:
                role_is_bound = self._binder.is_role_bound(role, bound_relationship.table)
                if role_is_bound:
                    # lookup referent entity maps
                    role_bindings = self._binder.lookup_role_bindings(role, bound_relationship.table)
                    role_maps = [entity_map for binding in role_bindings
                                 for entity_map in self._lookup_ref_entity_maps_by_column(role.player(), binding.column)]
                    for role_map in role_maps:
                        role_map_concept = role_map.value_player()
                        if role_map_concept is not role_concept and self._reasoner.in_subtype_closure(role_concept, role_map_concept):
                            role_map_casts_to[role_map] = role_concept
                else:
                    # lookup inferred entity map (note: must be unique?)
                    role_maps = [entity_map
                                 for entity_map in self._lookup_ref_entity_maps_by_source(role.player(), bound_relationship.table)]
            if not role_maps:
                raise ValueError(f'Cannot generate semantic predicate for {relationship},'
                                 f' no role maps found for role {role}')
            role_to_role_maps[role] = role_maps
        # if we got all roles with role maps, we can generate the rules
        role_map_combinations = [list(combination) for combination in product(*role_to_role_maps.values())]  # pyright: ignore[reportCallIssue]
        for role_maps in role_map_combinations:
            assert len(role_maps) > 0, "Must have at least one role map to derive a relationship"
            casted_var_refs, atoms = [], []
            for role_map in role_maps:
                var_ref, subformula = role_map.formula()
                if role_map in role_map_casts_to:
                    var_ref = role_map_casts_to[role_map](var_ref)
                casted_var_refs.append(var_ref)
                atoms.extend(subformula)
            where(*atoms).define(relationship(*casted_var_refs))

    def _try_materialize_entity_map(self, entity_map: AbstractInlineEntityMap) -> Union[AbstractInlineEntityMap, MaterializedEntityMap]:
        if self._cfg_should_inline_entity_maps():
            entity_map.materialize_population()
            return entity_map
        else:
            return entity_map.materialize()

    def _cfg_should_inline_value_maps(self):
        """
        Check if the config should inline value maps.
        """
        return self._config.get('inline_value_maps', DEFAULT_WEAVER_CONFIG['inline_value_maps'])

    def _cfg_should_inline_entity_maps(self):
        """
        Check if the config should inline entity maps.
        """
        return self._config.get('inline_entity_maps', DEFAULT_WEAVER_CONFIG['inline_entity_maps'])



from __future__ import annotations
from abc import ABC
from datetime import datetime
from typing import Any, Optional, TypedDict
from pathlib import Path
from urllib.parse import urlparse

from v0.relationalai.clients.util import IdentityParser

class Import(TypedDict):
    id: str
    name: str
    type: str
    model: str
    hash: Optional[str]
    info: Optional[dict]
    status: Optional[str]

class AvailableModel(TypedDict):
    name: Any
    id: Any
    created_by: Any
    created_on: Any
    state: Any
    deleted_by: Any
    deleted_on: Any

class EngineState(TypedDict):
    name: str
    id: str
    size: str
    state: str
    created_by: str
    created_on: datetime
    updated_on: datetime

    version: str|None
    auto_suspend: int|None
    suspends_at: datetime|None

class SourceInfo(TypedDict, total=False):
    type: str|None
    state: str
    columns_hash: str|None
    table_created_at: datetime|None
    stream_created_at: datetime|None
    last_ddl: datetime|None
    source: str


class ImportSource(ABC):
    name: str

class ImportSourceFile(ImportSource):
    def __init__(self, path: str, name: Optional[str] = None):
        self.raw_path = path
        self._url = urlparse(self.raw_path)
        self._path = Path(self._url.path if self.is_url() else self.raw_path)
        self.extension = self._path.suffix
        self.name = name or self._path.name

    def is_url(self):
        return bool(self._url.netloc)

class ImportSourceTable(ImportSource):
    def __init__(self, database: str, schema: str, table: str, name: Optional[str] = None):
        self.database = database
        self.schema = schema
        self.table = table
        self.name = name or self.fqn

    @property
    def fqn(self):
        full_name = f"{self.database}.{self.schema}.{self.table}"
        parser = IdentityParser(full_name)
        assert parser.identity, f"Failed to parse {full_name}"
        return parser.identity

class TransactionAsyncResponse:
    def __init__(
        self,
        transaction: dict|None = None,
        metadata: Any = None,
        results: list|None = None,
        problems: list|None = None,
    ):
        self.transaction = transaction
        self.metadata = metadata
        self.results = results
        self.problems = problems

    def __str__(self):
        return str(
            {
                "transaction": self.transaction,
                "metadata": self.metadata,
                "results": self.results,
                "problems": self.problems,
            }
        )

class ImportsStatus(TypedDict):
    engine: str
    engine_status: str
    engine_size: str
    status: str
    enabled: bool
    info: Any # @FIXME: This appears to be a stringified JSON (!?) object containing the same data as this dict + more

class SourceMapEntry(TypedDict):
    rel_end_line: int
    task_id: int
    py_line: int



from v0.relationalai.early_access.dsl.core.types.constrained import ConstrainedValueType
from v0.relationalai.early_access.dsl.core.types.variables import NominalTypeVariable


class ValueSubtype(ConstrainedValueType):

    def __init__(self, nm, *args):
        super().__init__(nm)
        for arg in args:
            self.build_scalar_variable([arg], {})
        self.name_type_variables()

    def arity(self):
        return self._parent.arity()

    def build_scalar_variable(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f"Unexpected error in ValueSubtype {self.display()}")

        if len(kwargs) != 0:
            raise Exception(
                f"Cannot use keyword arguments when instantiating a type parameter for ValueSubtype {self.display()}")

        ptype = args[0]
        if len(args) == 1:
            self._parent = ptype
            tvar = ptype.ranges_over()
            self._params = [tvar]
            return self.typevar_tuple()
        else:
            self._type_constraints.add_atomic_constraint(self.build_atom(ptype, args[1:]))

    def build_relation_variable(self, args, kwargs):
        if len(args) == 0:
            raise Exception(f"Unexpected error in ValueSubtype {self.display()}")

        if len(args) > 1:
            raise Exception(f"Cannot declare type variable of ValueSubtype {self.display()} by supplying more than a type")

        if len(kwargs) != 0:
            raise Exception(
                f"Cannot use keyword arguments when instantiating a type parameter for ValueSubtype {self.display()}")

        vtype = args[0]
        self._parent = vtype

        tvar = vtype.ranges_over()
        self._params.append(tvar)

        return tvar.typevar_tuple()

    def instantiate(self, *val):
        return self._parent.instantiate(val[0])

    def nominal(self):
        return self._parent.nominal()

    # Return a fresh type variable that ranges over this ValueSubtype
    #
    def ranges_over(self):

        # Self is a ValueSubtype, which is either nominal or not

        if self.nominal():
            # If nominal, then the variable that ranges over self must be a NominalTypeVariable
            tvar = NominalTypeVariable(self)
            tvar._parts = []
            vmap = {}
            selfparts = self._params[0]._parts
            for p in selfparts:
                (pprime, vmap) = p.dupe(vmap)
                tvar._parts.append(pprime)
            tvar._vmap = vmap

            # Because tvar might have multiple part variables, any subtype constraints that constrain
            # self will need to apply not to tvar.

            # Instantiate all scalar constraints from vt using vmap
            for c in self._type_constraints._sconstraints.values():
                cnew = c.revar(vmap)
                tvar._cset._sconstraints[id(cnew)] = cnew

            for c in self._type_constraints._atoms:
                cnew = c.revar(vmap)
                tvar._cset._atoms[id(cnew)] = cnew

            return tvar

        else:
            # If this type is not nominal, then it should have exactly one
            # TypeVariable parameter, which we can dupe and then extend
            # with any additional scalar constraints declared by this type.
            #
            lone_type_var = self._params[0]

            (tvar, vmap) = lone_type_var.dupe()
            vmap[id(lone_type_var)] = tvar
            # tvar._vmap = vmap
            tvar.decl = self
            for c in self._type_constraints._sconstraints.values():
                cnew = c.revar(vmap)
                tvar._cset._sconstraints[id(cnew)] = cnew
            return tvar

    def head_display(self):
        parent = self.parent()
        headvars = []
        for v in self._params:
            vt = v.typeof()
            if vt == parent:
                headvars.append(f"{v.display()}")
            else:
                headvars.append(f"{v.display()} in {vt.display()}")
        head = ", ".join(headvars)
        return f"value type {self.display()} <: {parent.display()}({head})"

    def parent(self):
        return self._parent

    def root_nominal_type(self):
        return self._parent.root_nominal_type()

    def root_unconstrained_type(self):
        return self._parent.root_unconstrained_type()
